"""This module contains extractor objects that perform extractions."""
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

import cv2  # type: ignore
import numpy as np
from moviepy.editor import VideoFileClip  # type: ignore

import videoxt.editors as E
import videoxt.requestors as R
from videoxt.exceptions import (
    AudioWriteError,
    BuildImagePathError,
    ClipWriteError,
    FrameReadError,
    FrameWriteError,
    GifWriteError,
    VideoCaptureSetError,
)
from videoxt.video import open_video_capture


class Extractor(Protocol):
    """Protocol defining the interface for performing any extraction operation."""

    def extract(self) -> Path | None:
        """
        Perform extraction and return the path to the extracted file or directory.

        Returns:
        -----
            `Path`: The path to the extracted file or directory.
        """
        ...


@dataclass
class AudioExtractor:  # XXX: Optimize.
    """
    Instantiate me with a `PreparedAudioRequest` and I'll provide you with an
    `extract()` method that will extract audio from the video file in the request,
    apply requested edits to the audio, and save it to disk.

    Fields:
    -----
        `request` (PreparedAudioRequest):
            Request parameters the extractor will use to perform the extraction.

    Public Methods:
    -----
        `extract()` -> `pathlib.Path`:
            Execute audio extraction and return the path to the extracted audio
            file.
    """

    request: R.PreparedAudioRequest

    def __post_init__(self) -> None:
        """Prepare the request if it has yet to be prepared."""
        self.request = (
            self.request
            if self.request.is_prepared
            else self.request.prepare()  # XXX: log
        )

    def extract(self) -> Path:
        """
        Extract audio from a video within a given time range and return the path to the
        saved audio file. Optional edits to audio are applied before saving to disk.

        Returns:
        -----
            `pathlib.Path | None`:
                The path to the extracted audio file if the write was successful.
        """
        with VideoFileClip(str(self.request.video.filepath)) as clip:
            subclip = self._edit_clip_audio(clip)
            return self._write_audio_file(subclip)

    def _edit_clip_audio(self, clip: VideoFileClip) -> VideoFileClip:
        """
        Apply optional edits to a clip before writing to disk as an audio file.

        Args:
        -----
            `clip` (moviepy.editor.VideoFileClip): The clip to apply edits to.

        Returns:
        -----
            `moviepy.editor.VideoFileClip`: The clip with edits applied.
        """
        start: float | None = self.request.extraction_range.get("start_second", None)
        stop: float | None = self.request.extraction_range.get("stop_second", None)
        clip = E.trim_clip(clip, start, stop)

        clip = E.edit_clip_motion(
            clip, self.request.speed, self.request.reverse, self.request.bounce
        )

        clip = E.edit_clip_audio(clip, self.request.volume, self.request.normalize)

        return clip

    def _write_audio_file(self, subclip: VideoFileClip) -> Path:
        """
        Write the audio from a subclip to disk and return the path to the audio file.

        Returns:
        -----
            `pathlib.Path`:
                The path to the extracted audio file if the write was successful.

        Raises:
        -----
            `AudioWriteError`:
                If the audio file could not be written to disk (from OSError).
        """
        try:
            subclip.audio.write_audiofile(
                str(self.request.destpath), logger=None, fps=44100
            )
        except OSError as e:
            raise AudioWriteError(
                f"Error writing audio to {self.request.destpath}"
            ) from e
        else:
            return self.request.destpath


@dataclass
class ClipExtractor:  # XXX: Optimize.
    """
    Instantiate me with a `PreparedClipRequest` and I'll provide you with an `extract()`
    method that will extract a subclip from the video file in the request, apply
    requested edits to the subclip, and save it to disk.

    Fields:
    -----
        `request` (PreparedClipRequest):
            Request parameters the extractor will use to perform the extraction.

    Public Methods:
    -----
        `extract()` -> `pathlib.Path`:
            Execute subclip extraction and return the path to the extracted clip.
    """

    request: R.PreparedClipRequest

    def __post_init__(self) -> None:
        """Prepare the request if it has yet to be prepared."""
        self.request = (
            self.request
            if self.request.is_prepared
            else self.request.prepare()  # XXX: log
        )

    def extract(self) -> Path:
        """
        Extract a subclip from a video within a given time range and return the file
        path to clip. Optional edits to the clip are applied before saving to disk.

        Returns:
        -----
            `pathlib.Path`:
                The path to the extracted clip if the write was successful.
        """
        with VideoFileClip(str(self.request.video.filepath)) as clip:
            subclip = self._edit_clip(clip)
            return self._write_subclip(subclip)

    def _edit_clip(self, clip: VideoFileClip) -> VideoFileClip:
        """
        Apply optional edits to a clip before writing to disk as a 'mp4' file.

        Args:
        -----
            `clip` (moviepy.editor.VideoFileClip): The clip to edit.

        Returns:
        -----
            `moviepy.editor.VideoFileClip`: The edited clip.
        """
        start: float | None = self.request.extraction_range.get("start_second", None)
        stop: float | None = self.request.extraction_range.get("stop_second", None)
        clip = E.trim_clip(clip, start, stop)

        clip = E.edit_clip_audio(clip, self.request.volume, self.request.normalize)

        clip = E.edit_clip_motion(
            clip, self.request.speed, self.request.reverse, self.request.bounce
        )

        clip = E.edit_clip_image(
            clip, self.request.dimensions, self.request.rotate, self.request.monochrome
        )

        return clip

    def _write_subclip(self, subclip: VideoFileClip) -> Path:
        """
        Write a subclip to disk and return the path to the clip.

        Returns:
        -----
            `pathlib.Path`:
                The path to the extracted clip if the write was successful.

        Raises:
        -----
            `ClipWriteError`: If the clip could not be written to disk (from OSError).
        """
        try:
            subclip.write_videofile(str(self.request.destpath), logger=None)
        except OSError as e:
            raise ClipWriteError(
                f"Error writing clip to {self.request.destpath}"
            ) from e
        else:
            return self.request.destpath


@dataclass
class FramesExtractor:  # XXX: Optimize.
    """
    Instantiate me with a `PreparedFramesRequest` and I'll provide you with an
    `extract()` method that will extract frames from the video file in the request,
    apply requested edits to the frames, and save them to disk as images in a directory.

    Fields:
    -----
        `request` (PreparedFramesRequest):
            The request to prepare and perform an extraction operation with.

    Public Methods:
    -----
        `extract()` -> `pathlib.Path`:
            Execute frames extraction and return the path to the directory where the
            extracted images were saved.
    """

    request: R.PreparedFramesRequest

    def __post_init__(self) -> None:
        """Prepare the request if it has yet to be prepared."""
        self.request = (
            self.request
            if self.request.is_prepared
            else self.request.prepare()  # XXX: log
        )

    def extract(self) -> Path:
        """
        Extract frames from a video, save them to disk as images, and return the path
        to the directory where the images were saved.

        Returns:
        -----
            `pathlib.Path`: The directory where the extracted images were saved.
        """

        # Create the destination directory if it doesn't exist.
        self.request.destpath.mkdir(parents=True, exist_ok=True)

        from rich.progress import track

        # Open the video capture and iterate over the frames to write to disk.
        with open_video_capture(self.request.video.filepath) as opencap:
            for edited_frame, image_path in track(
                self._preprocess_frames(opencap),
                total=self.request.images_expected,
                transient=True,
                description=(
                    "[yellow]Extracting frames...[/yellow]\n"
                    "Press [red][bold]Ctrl+C[/red][/bold] to cancel."
                ),
            ):
                self._write_image(edited_frame, image_path)

        return self.request.destpath

    def _preprocess_frames(
        self, opencap: cv2.VideoCapture
    ) -> Generator[tuple[np.ndarray, Path], None, None]:
        """
        Edit the frames to be extracted and yield the image path and the edited frame.

        Args:
        -----
            `opencap` (cv2.VideoCapture): The open video capture to read from.

        Yields:
        -----
            `tuple[pathlib.Path, np.ndarray]`:
                A tuple containing the image path and frame.
        """
        # Iterate over the range of frames to extract.
        for frame_num in self._generate_frame_numbers():
            # Build the image path, read video frame, apply edits, yield path and frame.
            frame = self._read_video_frame(opencap, frame_num)
            edited_frame = self._edit_video_frame(frame)
            image_path = self._build_image_path(frame_num)
            yield edited_frame, image_path

    def _generate_frame_numbers(self) -> Generator[int, None, None]:
        start_frame = self.request.extraction_range.get("start_frame", 0)
        stop_frame = self.request.extraction_range.get(
            "stop_frame", self.request.video.frame_count
        )
        capture_rate = self.request.capture_rate
        images_expected = self.request.images_expected

        current_frame = start_frame
        while current_frame < stop_frame and images_expected > 0:
            yield current_frame
            current_frame += capture_rate
            images_expected -= 1

    def _build_image_path(self, frame_num: int) -> Path:
        """
        Build the filepath for the next image to be written.

        Args:
        -----
            `frame_num` (int):The frame number to build the filepath for.

        Returns:
        -----
            `pathlib.Path`: The filepath for the next image to be written.

        Raises:
        -----
            `BuildImagePathError`:
                If the destination directory or filename is None.
        """
        if self.request.destpath is None or self.request.filename is None:
            raise BuildImagePathError("The destination directory or filename is None.")

        return (
            self.request.destpath
            / f"{self.request.filename}_{frame_num+1}.{self.request.image_format}"
        )

    def _read_video_frame(
        self, opencap: cv2.VideoCapture, frame_num: int
    ) -> np.ndarray:
        """
        Point the video capture to the frame number to read (0-based index) and return
        a `np.ndarray` representing the video frame.

        Args:
        -----
            `opencap` (cv2.VideoCapture):
                The open video capture to read from.
            `frame_num` (int):
                The frame number to read.

        Returns:
        -----
            `np.ndarray`: The video frame from the capture.

        Raises:
        -----
            `VideoCaptureSetError`:
                If opencap.set(cv2.CAP_PROP_POS_FRAMES, frame_num) returns False.
            `FrameReadError`:
                If opencap.read() could not grab, decode and return the frame.

        Notes:
        -----
            - set() uses a 0-based index of the frame to be decoded/captured next.
            - set() returns True if the capture device has accepts the property value,
            even if property value remains unchanged.
            - read() returns a tuple of (bool, np.ndarray).
        """
        set_successful = opencap.set(cv2.CAP_PROP_POS_FRAMES, frame_num)
        if not set_successful:
            raise VideoCaptureSetError(
                "Could not set 'cv2.CAP_PROP_POS_FRAMES' property in the video "
                f"capture to frame number {frame_num}."
            )

        read_successful, frame = opencap.read()
        if not read_successful:
            raise FrameReadError(
                f"Could not read frame {frame_num} from video capture."
            )

        return frame

    def _edit_video_frame(self, frame: np.ndarray) -> np.ndarray:
        """
        Apply optional edits to a np.ndarray before writing to disk as an image.

        Args:
        -----
            `frame` (np.ndarray): The video frame to edit.

        Returns:
        -----
            `np.ndarray`: The edited video frame.
        """
        return E.edit_image(
            frame, self.request.dimensions, self.request.rotate, self.request.monochrome
        )

    def _write_image(self, edited_frame: np.ndarray, image_path: Path) -> None:
        """
        Write an edited frame to disk as an image.

        Args:
        -----
            `edited_frame` (np.ndarray):
                The edited video frame to write to disk as an image.
            `image_path` (pathlib.Path):
                The path to write the image to.

        Raises:
        -----
            `FrameWriteError`:
            - If cv2.imwrite() raises an OSError.
            - If cv2.imwrite() returns False.
        """
        try:
            write_successful = cv2.imwrite(str(image_path), edited_frame)
        except OSError as e:
            raise FrameWriteError(f"Error writing image to {image_path}") from e
        else:
            if not write_successful:
                raise FrameWriteError(f"Unsuccessful write to {image_path}")


@dataclass
class GifExtractor:  # XXX: Optimize.
    """
    Instantiate me with a `PreparedGifRequest` and I'll provide you with an `extract()`
    method that will extract a gif from the video file in the request, apply requested
    edits to the gif, and save it to disk.

    Attributes:
    -----
        `request` (PreparedGifRequest):
            Request parameters the extractor will use to perform the extraction.

    Public Methods:
    -----
        `extract()` -> `pathlib.Path`:
            Execute gif extraction and return the path to the extracted gif.
    """

    request: R.PreparedGifRequest

    def __post_init__(self) -> None:
        """Prepare the request if it has yet to be prepared."""
        self.request = (
            self.request
            if self.request.is_prepared
            else self.request.prepare()  # XXX: log
        )

    def extract(self) -> Path:
        """
        Extract a gif from a video within a given time range and return the file path
        the gif. Optional edits to the gif are applied before saving to disk.

        Returns:
        -----
            `pathlib.Path`:
                The path to the extracted gif if the write was successful.
        """
        with VideoFileClip(str(self.request.video.filepath)) as clip:
            subclip = self._edit_clip(clip)
            return self._write_gif(subclip)

    def _edit_clip(self, clip: VideoFileClip) -> VideoFileClip:
        """
        Apply optional edits to a clip.

        Args:
        -----
            `clip` (moviepy.editor.VideoFileClip): The clip to edit.

        Returns:
        -----
            `moviepy.editor.VideoFileClip`: The edited clip.
        """
        start: float | None = self.request.extraction_range.get("start_second", None)
        stop: float | None = self.request.extraction_range.get("stop_second", None)
        clip = E.trim_clip(clip, start, stop)

        clip = E.edit_clip_image(
            clip, self.request.dimensions, self.request.rotate, self.request.monochrome
        )

        clip = E.edit_clip_motion(
            clip, self.request.speed, self.request.reverse, self.request.bounce
        )

        return clip

    def _write_gif(self, subclip: VideoFileClip) -> Path:
        """
        Write an edited subclip to disk as a gif.

        Returns:
        -----
            `pathlib.Path`:
                The path to the extracted gif if the write was successful.

        Raises:
        -----
            `GifWriteError`: If the gif could not be written to disk (from OSError).
        """
        try:
            subclip.write_gif(str(self.request.destpath), logger=None)
        except OSError as e:
            raise GifWriteError(
                f"Error writing gif to {self.request.destpath!r}"
            ) from e
        else:
            return self.request.destpath



import shutil
from collections.abc import Generator
from pathlib import Path
from typing import Any

import pytest

from videoxt.api import extract_audio, extract_clip, extract_frames, extract_gif
from videoxt.result import Result


@pytest.fixture(scope="session")
def extracted_frames(fixture_tmp_video_filepath) -> Generator[Result, None, None]:
    """Extract frames from a video and yield a Result object."""
    result = extract_frames(fixture_tmp_video_filepath)
    yield result
    if result.destpath is not None:
        shutil.rmtree(Path(result.destpath))


def test_extract_frames_valid_default_request_success_is_true(
    extracted_frames: Result, fixture_tmp_video_properties: dict[str, Any]
):
    assert extracted_frames.success is True
    assert extracted_frames.method == "frames"
    assert extracted_frames.message == "Extraction successful."
    expected_destpath = (
        fixture_tmp_video_properties["video_file_path"].parent / "tmp.video.mp4_frames"
    )
    assert extracted_frames.destpath == expected_destpath
    expected_destpath = Path(expected_destpath)
    assert expected_destpath.exists()
    assert expected_destpath.is_dir()


@pytest.fixture(scope="session")
def extracted_audio(fixture_tmp_video_filepath) -> Generator[Result, None, None]:
    """Extract audio from a video and yield a Result object."""
    result = extract_audio(fixture_tmp_video_filepath)
    yield result
    if result.destpath is not None:
        Path(result.destpath).unlink()


def test_extract_audio_valid_default_request_success_is_true(
    extracted_audio: Result, fixture_tmp_video_properties: dict[str, Any]
):
    assert extracted_audio.success is True
    assert extracted_audio.method == "audio"
    assert extracted_audio.message == "Extraction successful."
    expected_destpath = (
        fixture_tmp_video_properties["video_file_path"].parent / "tmp.video.mp3"
    )
    assert extracted_audio.destpath == expected_destpath
    expected_destpath = Path(expected_destpath)
    assert expected_destpath.exists()
    assert expected_destpath.is_file()


@pytest.fixture(scope="session")
def extracted_clip(fixture_tmp_video_filepath) -> Generator[Result, None, None]:
    """Extract a clip from a video and yield a Result object."""
    result = extract_clip(fixture_tmp_video_filepath, filename="tmp.test.extract.clip")
    yield result
    if result.destpath is not None:
        Path(result.destpath).unlink()


def test_extract_clip_valid_default_request_success_is_true(
    extracted_clip: Result, fixture_tmp_video_properties: dict[str, Any]
):
    assert extracted_clip.success is True
    assert extracted_clip.method == "clip"
    assert extracted_clip.message == "Extraction successful."
    expected_destpath = (
        fixture_tmp_video_properties["video_file_path"].parent
        / "tmp.test.extract.clip.mp4"
    )
    assert extracted_clip.destpath == expected_destpath
    expected_destpath = Path(expected_destpath)
    assert expected_destpath.exists()
    assert expected_destpath.is_file()


@pytest.fixture(scope="session")
def extracted_gif(fixture_tmp_video_filepath) -> Generator[Result, None, None]:
    """Extract a gif from a video and yield a Result object."""
    result = extract_gif(fixture_tmp_video_filepath)
    yield result
    if result.destpath is not None:
        Path(result.destpath).unlink()


def test_extract_gif_valid_default_request_success_is_true(
    extracted_gif: Result, fixture_tmp_video_properties: dict[str, Any]
):
    assert extracted_gif.success is True
    assert extracted_gif.method == "gif"
    assert extracted_gif.message == "Extraction successful."
    expected_destpath = (
        fixture_tmp_video_properties["video_file_path"].parent / "tmp.video.gif"
    )
    assert extracted_gif.destpath == expected_destpath
    expected_destpath = Path(expected_destpath)
    assert expected_destpath.exists()
    assert expected_destpath.is_file()



"""Pytest configuration file containing shared fixtures."""
import os
import shutil
from collections.abc import Generator
from datetime import timedelta
from pathlib import Path
from typing import Any

import cv2  # type: ignore
import numpy as np
import pytest
from moviepy.editor import AudioClip, VideoClip  # type: ignore


@pytest.fixture(scope="session")
def fixture_tmp_dir() -> Generator[Path, None, None]:
    """
    Create a temporary directory for the session and yield its path.

    Yields:
    -----
        `Generator[Path, None, None]`:
            A generator that yields the path to the temporary directory.
    """
    temp_dir = Path("tmp")
    temp_dir.mkdir(exist_ok=True)
    yield temp_dir
    shutil.rmtree(temp_dir)


@pytest.fixture(scope="session")
def fixture_tmp_video_properties(fixture_tmp_dir: Path) -> dict[str, Any]:
    """
    Returns a dictionary of video properties containing values used to create a valid
    temporary video file in `tmp_video_filepath`, and used as expected values for tests.

    Args:
    -----
        `fixture_tmp_dir` (pathlib.Path):
            A fixture that creates a temporary directory for the session.

    Returns:
    -----
        `dict[str, Any]`: A dictionary of properties for the temporary video.
    """
    return {
        "video_file_path": fixture_tmp_dir / "tmp.video.mp4",
        "filename": "tmp.video",
        "audio_codec": "aac",
        "codec": "libx264",
        "dimensions": (640, 480),
        "duration": timedelta(seconds=2),
        "duration_seconds": 2,
        "duration_timestamp": "0:00:02",
        "fps": 10.0,
        "frame_count": 20,
        "frame_height": 480,
        "frame_width": 640,
        "has_audio": True,
    }


@pytest.fixture(scope="session")
def fixture_tmp_video_filepath(
    fixture_tmp_video_properties: dict[str, Any]
) -> Generator[Path, None, None]:
    """
    Create a temporary video file and yield its path.

    Args:
    -----
        `fixture_tmp_video_properties` (dict[str, Any]):
            A fixture that returns a dictionary of properties for the temporary video.

    Yields:
    -----
        `Generator[Path, None, None]`:
            A generator that yields the path to the temporary video file.
    """

    def make_frame(time: float) -> np.ndarray:
        """Generate mock frames for the length of the video."""
        frame_num = int(time * fixture_tmp_video_properties["fps"]) + 1
        frame = np.zeros(
            (
                fixture_tmp_video_properties["frame_height"],
                fixture_tmp_video_properties["frame_width"],
                3,
            ),
            dtype=np.uint8,
        )
        text = f"Frame: {frame_num}"
        cv2.putText(frame, text, (50, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
        return frame

    def make_audio(time: float) -> float:
        """Generate mock audio for the length of the video."""
        return np.sin(2 * np.pi * 440 * time)

    # Generate the video and set the audio
    video = VideoClip(
        make_frame, duration=fixture_tmp_video_properties["duration_seconds"]
    )
    audio = AudioClip(
        make_audio, duration=fixture_tmp_video_properties["duration_seconds"]
    )
    video = video.set_audio(audio)

    # Write the video to disk
    video.write_videofile(
        str(fixture_tmp_video_properties["video_file_path"]),
        codec=fixture_tmp_video_properties["codec"],
        audio_codec=fixture_tmp_video_properties["audio_codec"],
        fps=fixture_tmp_video_properties["fps"],
    )

    # Yield the filepath of the temporary video file then delete it
    yield fixture_tmp_video_properties["video_file_path"]

    # Delete the temporary video file
    if fixture_tmp_video_properties["video_file_path"].exists():
        os.remove(fixture_tmp_video_properties["video_file_path"])


@pytest.fixture(scope="session")
def fixture_tmp_video_filepath_zero_seconds() -> Generator[Path, None, None]:
    """
    Generate a temporary 'mp4' video file and provide its path. This file is invalid
    due to its 0-second duration, absence of audio, and lack of frames.

    Yields:
    -----
        `Generator[Path, None, None]`:
            A generator that yields the path to the temporary video file.
    """
    video_file_path = Path("tmp") / "tmp.invalid.video.mp4"

    # Generate the video and set the audio
    video = VideoClip(lambda t: np.zeros((480, 640, 3), dtype=np.uint8), duration=0)
    video.write_videofile(str(video_file_path), fps=10)

    yield video_file_path

    # Delete the temporary video file
    if video_file_path.exists():
        os.remove(video_file_path)


@pytest.fixture(scope="session")
def fixture_tmp_text_filepath(fixture_tmp_dir: Path) -> Generator[Path, None, None]:
    """
    Create a temporary text file in the temporary directory and yield its path.

    Yields:
    -----
        `Generator[Path, None, None]`:
            A generator that yields the path to the temporary text file.
    """
    tmp_filepath = fixture_tmp_dir / "tmp.text.txt"
    content = "This is a temporary text file."
    with open(tmp_filepath, "w") as f:
        f.write(content)

    yield tmp_filepath

    if tmp_filepath.exists():
        os.remove(tmp_filepath)



"""
Contains factories for creating extraction objects and handlers for executing the
extraction process.
"""
from pathlib import Path
from time import perf_counter
from typing import Any, Optional

from rich.console import Console

from videoxt.constants import ExtractionMethod
from videoxt.exceptions import VideoXTError
from videoxt.extractors import (
    AudioExtractor,
    ClipExtractor,
    Extractor,
    FramesExtractor,
    GifExtractor,
)
from videoxt.requestors import (
    AudioRequest,
    ClipRequest,
    FramesRequest,
    GifRequest,
    PreparedAudioRequest,
    PreparedClipRequest,
    PreparedFramesRequest,
    PreparedGifRequest,
    PreparedRequest,
    Request,
)
from videoxt.result import Result
from videoxt.video import Video


class ExtractionHandler:
    """
    Handles the extraction process by creating the necessary objects used to perform
    any extraction.

    Attributes:
    -----
        `method` (ExtractionMethod):
            The extraction method Enum to use (e.g., `ExtractionMethod.AUDIO`).

    Public Methods:
    -----
        `execute` -> `Result`:
            Execute the video extraction process and return a `Result` object.
    """

    def __init__(self, method: ExtractionMethod):
        """Set the objects to be used for extraction based on the chosen method."""
        self.method = method
        self.object_factory = ObjectFactory(self.method)

    def execute(
        self,
        filepath: Path | str,
        options: Optional[dict[str, Any]] = None,
        skip_validation: bool = False,
    ) -> Result:
        """
        Make the necessary objects used to perform the extraction and execute the
        extraction process.

        Args:
        -----
            `filepath` (Path | str):
                Path to the video file with extension.
            `options` (Optional[dict[str, Any]]):
                Extraction options specific to the chosen extraction method. If None,
                default options will be used.
            `skip_validation` (bool):
                If True, skips validation of the extraction options. This can slightly
                improve speed, but it is not recommended unless you are sure that the
                options are valid.
        Returns:
        -----
            `Result`: A dataclass containing the extraction details.
        """
        video = self.object_factory.make_video(filepath)
        request = self.object_factory.make_prepared_request(
            video, options, skip_validation
        )
        extractor = self.object_factory.make_extractor(request)
        result = self.object_factory.make_result()

        if request.is_verbose:
            request.verbose_print(title="PreparedRequest")

        result = self._perform_extraction(extractor, result)

        if request.is_verbose:
            result.verbose_print(title="Result")

        return result

    def _perform_extraction(self, extractor: Extractor, result: Result) -> Result:
        """
        Execute the extraction process and return a `Result` object.

        Args:
        -----
            `extractor` (Extractor):
                The extractor object to use.
            `result` (Result):
                The result object to return.

        Returns:
        -----
            `Result`: A dataclass containing the extraction details.
        """
        _timer_start = perf_counter()
        try:
            if self.method == ExtractionMethod.FRAMES:
                extractor.extract()
            else:
                transient_msg = (
                    f"[yellow]Extracting {self.method.value}...[/yellow]\n"
                    "Press [red][bold]Ctrl+C[/red][/bold] to cancel."
                )
                with Console().status(transient_msg):
                    extractor.extract()

        except KeyboardInterrupt:
            result.success = False
            result.message = "Extraction cancelled."

        except VideoXTError as error_msg:
            result.success = False
            result.message = f"Extraction failed: {error_msg}"

        else:
            result.success = True
            result.message = "Extraction successful."

        finally:
            if extractor.request.destpath.exists():
                result.destpath = extractor.request.destpath

            result.elapsed_time = perf_counter() - _timer_start
            return result


class ObjectFactory:
    """
    Factory for creating the necessary objects used to perform any extraction.

    Attributes:
    -----
        `method` (ExtractionMethod):
            The extraction method Enum to use (e.g., `ExtractionMethod.AUDIO`).

    Class Attributes:
    -----
        `REQUEST_MAP` (dict[ExtractionMethod, Type[Request]]):
            Mapping of extraction methods to their corresponding request classes.
        `PREPARED_REQUEST_MAP` (dict[ExtractionMethod, Type[PreparedRequest]]):
            Mapping of extraction methods to their corresponding prepared request
            classes.
        `EXTRACTOR_MAP` (dict[ExtractionMethod, Type[Extractor]]):
            Mapping of extraction methods to their corresponding extractor classes.

    Public Methods:
    -----
        `make_video` -> `Video`:
            Create a `Video` object.
        `make_prepared_request` -> `PreparedRequest`:
            Create a `PreparedRequest` object.
        `make_extractor` -> `Extractor`:
            Create an `Extractor` object.
        `make_result` -> `Result`:
            Create a `Result` object.
    """

    REQUEST_MAP: dict[ExtractionMethod, type[Request]] = {
        ExtractionMethod.AUDIO: AudioRequest,
        ExtractionMethod.CLIP: ClipRequest,
        ExtractionMethod.FRAMES: FramesRequest,
        ExtractionMethod.GIF: GifRequest,
    }
    PREPARED_REQUEST_MAP: dict[ExtractionMethod, type[PreparedRequest]] = {
        ExtractionMethod.AUDIO: PreparedAudioRequest,
        ExtractionMethod.CLIP: PreparedClipRequest,
        ExtractionMethod.FRAMES: PreparedFramesRequest,
        ExtractionMethod.GIF: PreparedGifRequest,
    }
    EXTRACTOR_MAP: dict[ExtractionMethod, type[Extractor]] = {
        ExtractionMethod.AUDIO: AudioExtractor,
        ExtractionMethod.CLIP: ClipExtractor,
        ExtractionMethod.FRAMES: FramesExtractor,
        ExtractionMethod.GIF: GifExtractor,
    }

    def __init__(self, method: ExtractionMethod):
        """Set the objects to be used for extraction based on the chosen method."""
        self.method = method

    def make_video(self, filepath: Path | str) -> Video:
        """
        Use the given video filepath to create a `Video` object.

        Args:
        -----
            `filepath` (Path | str): Path to the video file with extension.

        Returns:
        -----
            `Video`: A dataclass containing the video details.
        """
        return Video(Path(filepath))

    def make_prepared_request(
        self,
        video: Video,
        options: Optional[dict[str, Any]] = None,
        skip_validation: bool = False,
    ) -> PreparedRequest:
        """
        Use the given video and options to create a `PreparedRequest` object.

        Args:
        -----
            `video` (Video):
                The video object to use.
            `options` (Optional[dict[str, Any]]):
                Extraction options specific to the chosen extraction method. If None,
                default options will be used.
            `skip_validation` (bool):
                If True, skips validation of the extraction options. This can slightly
                improve speed, but it is not recommended unless you are sure that the
                options are valid.

        Returns:
        -----
            `PreparedRequest`: A dataclass containing the prepared request details.
        """
        if options is None:
            return self.PREPARED_REQUEST_MAP[self.method](video).prepare()

        if skip_validation:
            return self.PREPARED_REQUEST_MAP[self.method](video, **options).prepare()

        return self.REQUEST_MAP[self.method](**options).validate().prepare(video)

    def make_extractor(self, prepared_request: PreparedRequest) -> Extractor:
        """
        Use the given prepared request to create an `Extractor` object.

        Args:
        -----
            `prepared_request` (PreparedRequest): A type of prepared request to use.

        Returns:
        -----
            `Extractor`: A type of extractor object.
        """
        return self.EXTRACTOR_MAP[self.method](prepared_request)

    def make_result(self) -> Result:
        """Create and return a `Result` object with the chosen extraction method."""
        return Result(method=self.method.value)



"""Simple interface for extracting audio, clips, frames, and gifs from a video file."""
from pathlib import Path
from typing import Any, Optional

from videoxt.constants import ExtractionMethod
from videoxt.exceptions import InvalidExtractionMethod
from videoxt.handlers import ExtractionHandler
from videoxt.result import Result


def extract(
    method: str,
    filepath: Path | str,
    skip_validation: bool = False,
    **options: Optional[dict[str, Any]],
) -> Result:
    """
    Extract audio, individual frames, short clips and GIFs from videos.

    See the documentation for list of options, or:
    - `extract_audio`: Extract audio from a video file.
    - `extract_clip`: Extract a short clip from a video file as `mp4`.
    - `extract_frames`: Extract individual frames from a video and save them as images.
    - `extract_gif`: Create a GIF from a video between two specified points.

    Args:
    -----
        `method` (str):
            The extraction method to use ("audio", "clip", "frames", "gif").
        `filepath` (pathlib.Path | str):
            Path to the video file with extension.
        `skip_validation` (bool):
            If True, skips validation of the extraction options. This can slightly
            improve speed, but it is not recommended unless you are sure that the
            options are valid.
        `**options` (Optional[dict[str, Any]]):
            Extraction options specific to the chosen extraction method.

    Returns:
    -----
        `Result`: A dataclass containing the extraction details.

    Raises:
    -----
        `InvalidExtractionMethod`:
            If the extraction method is neither "audio", "clip", "frames", nor "gif".
    """
    try:
        method_enum = ExtractionMethod[method.upper()]

    except KeyError:
        raise InvalidExtractionMethod(
            f"Invalid extraction method: {method}. "
            f"Choices are {', '.join([em.value for em in ExtractionMethod])}."
        )

    else:
        handler = ExtractionHandler(method_enum)
        return handler.execute(filepath, options, skip_validation=skip_validation)


def extract_audio(
    filepath: Path | str,
    start_time: float | int | str = 0,
    stop_time: Optional[float | int | str] = None,
    destdir: Optional[Path | str] = None,
    filename: Optional[str] = None,
    verbose: bool = False,
    overwrite: bool = False,
    fps: Optional[float] = None,
    audio_format: str = "mp3",
    speed: float = 1,
    bounce: bool = False,
    reverse: bool = False,
    volume: float = 1,
    normalize: bool = False,
) -> Result:
    """
    Extract audio from a video file.

    Args:
    -----
        `filepath` (pathlib.Path | str):
            Path to the video file with extension.
        `start_time` (float | int | str):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (bool):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (bool):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `audio_format` (str):
            Set the extracted audio file format. Defaults to 'mp3' if not specified.
            See: `videoxt.constants.SUPPORTED_AUDIO_FORMATS`.
        `speed` (float):
            Set the speed of the extracted audio. A value of 0.5 will halve the speed of
            the extracted audio. Defaults to 1.0 if not specified (no change).
        `bounce` (bool):
            If True, bounce the extracted audio bommerang-style. Defaults to False if
            not specified.
        `reverse` (bool):
            If True, reverse the extracted audio. Defaults to False if not specified.
        `volume` (float):
            Set the volume of the extracted audio. A value of 0.5 will halve the volume
            of the extracted audio. Defaults to 1.0 if not specified (no change).
        `normalize` (bool):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Returns:
    -----
        `Result`: A dataclass containing the extraction details.
    """
    options = {
        "start_time": start_time,
        "stop_time": stop_time,
        "destdir": destdir,
        "filename": filename,
        "verbose": verbose,
        "overwrite": overwrite,
        "fps": fps,
        "audio_format": audio_format,
        "speed": speed,
        "bounce": bounce,
        "reverse": reverse,
        "volume": volume,
        "normalize": normalize,
    }

    handler = ExtractionHandler(ExtractionMethod.AUDIO)
    return handler.execute(filepath, options)


def extract_clip(
    filepath: Path | str,
    start_time: float | int | str = 0,
    stop_time: Optional[float | int | str] = None,
    destdir: Optional[Path | str] = None,
    filename: Optional[str] = None,
    verbose: bool = False,
    overwrite: bool = False,
    fps: Optional[float] = None,
    dimensions: Optional[tuple[int, int]] = None,
    resize: float = 1.0,
    rotate: int = 0,
    speed: float = 1.0,
    bounce: bool = False,
    reverse: bool = False,
    monochrome: bool = False,
    volume: float = 0,
    normalize: bool = False,
) -> Result:
    """
    Extract a clip from a video file as `mp4`.

    Recommended usage: Set a short extraction range. The process can be slow for long
    or high-resolution videos.

    Args:
    -----
        `filepath` (pathlib.Path | str):
            Path to the video file with extension.
        `start_time` (float | int | str):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (bool):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (bool):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the clip. Defaults to
            the video dimensions if not specified.
        `resize` (float):
            Resize the dimensions of the clip by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (int):
            Rotate the clip by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (float):
            Set the speed of the extracted clip. A value of 0.5 will halve the playback
            speed of the clip. Defaults to 1.0 if not specified (no change).
        `bounce` (bool):
            If True, bounce the extracted clip bommerang-style. Defaults to False if
            not specified.
        `reverse` (bool):
            If True, reverse the extracted clip. Defaults to False if not specified.
        `monochrome` (bool):
            If True, apply a black-and-white filter to the clip. Defaults to False if
            not specified.
        `volume` (float):
            Set the volume of the extracted clip's audio. A value of 0.5 will halve the
            volume of the clip's audio. Defaults to 1.0 if not specified (no change).
        `normalize` (bool):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Returns:
    -----
        `Result`: A dataclass containing the extraction details.
    """
    options = {
        "start_time": start_time,
        "stop_time": stop_time,
        "destdir": destdir,
        "filename": filename,
        "verbose": verbose,
        "overwrite": overwrite,
        "fps": fps,
        "dimensions": dimensions,
        "resize": resize,
        "rotate": rotate,
        "speed": speed,
        "bounce": bounce,
        "reverse": reverse,
        "monochrome": monochrome,
        "volume": volume,
        "normalize": normalize,
    }

    handler = ExtractionHandler(ExtractionMethod.CLIP)
    return handler.execute(filepath, options)


def extract_frames(
    filepath: Path | str,
    start_time: float | int | str = 0,
    stop_time: Optional[float | int | str] = None,
    destdir: Optional[Path | str] = None,
    filename: Optional[str] = None,
    verbose: bool = False,
    overwrite: bool = False,
    fps: Optional[float] = None,
    image_format: str = "jpg",
    capture_rate: int = 1,
    dimensions: Optional[tuple[int, int]] = None,
    resize: float = 1.0,
    rotate: int = 0,
    monochrome: bool = False,
) -> Result:
    """
    Extract individual frames from a video and save them to disk as images.

    The images are saved to a directory named after the video file, or to a directory
    you specify.

    Args:
    -----
        `filepath` (pathlib.Path | str):
            Path to the video file with extension.
        `start_time` (float | int | str):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (bool):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (bool):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `image_format` (str):
            Set the extracted image file format. Defaults to 'jpg' if not specified.
            See: `videoxt.constants.SUPPORTED_IMAGE_FORMATS`.
        `capture_rate` (int):
            Capture every Nth video frame. Defaults to 1 if not specified, which
            extracts every frame within the extraction range.
        `dimensions` (Optional[tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the images. Defaults
            to the video dimensions if not specified.
        `resize` (float):
            Resize the dimensions of the images by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (int):
            Rotate the images by `n` degrees. Allowed values: 0, 90, 180 or 270.
            Defaults to 0 if not specified (no change).
        `monochrome` (bool):
            If True, apply a black-and-white filter to the images. Defaults to False if
            not specified.

    Returns:
    -----
        `Result`: A dataclass containing the extraction details.
    """
    options = {
        "start_time": start_time,
        "stop_time": stop_time,
        "destdir": destdir,
        "filename": filename,
        "verbose": verbose,
        "overwrite": overwrite,
        "fps": fps,
        "image_format": image_format,
        "capture_rate": capture_rate,
        "dimensions": dimensions,
        "resize": resize,
        "rotate": rotate,
        "monochrome": monochrome,
    }

    handler = ExtractionHandler(ExtractionMethod.FRAMES)
    return handler.execute(filepath, options)


def extract_gif(
    filepath: Path | str,
    start_time: float | int | str = 0,
    stop_time: Optional[float | int | str] = None,
    destdir: Optional[Path | str] = None,
    filename: Optional[str] = None,
    verbose: bool = False,
    overwrite: bool = False,
    fps: Optional[float] = None,
    dimensions: Optional[tuple[int, int]] = None,
    resize: float = 1.0,
    rotate: int = 0,
    speed: float = 1.0,
    bounce: bool = False,
    reverse: bool = False,
    monochrome: bool = False,
) -> Result:
    """
    Extract a GIF from a video file.

    Recommended usage: Set a short extraction range. The process can be slow for long
    or high-resolution videos.

    Args:
    -----
        `filepath` (pathlib.Path | str):
            Path to the video file with extension.
        `start_time` (float | int | str):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (bool):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (bool):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the gif. Defaults to
            the video dimensions if not specified.
        `resize` (float):
            Resize the dimensions of the gif by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (int):
            Rotate the gif by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (float):
            Set the speed of the extracted gif. A value of 0.5 will halve the playback
            speed of the gif. Defaults to 1.0 if not specified (no change).
        `bounce` (bool):
            If True, bounce the extracted gif bommerang-style. Defaults to False if
            not specified.
        `reverse` (bool):
            If True, reverse the extracted gif. Defaults to False if not specified.
        `monochrome` (bool):
            If True, apply a black-and-white filter to the gif. Defaults to False if
            not specified.

    Returns:
    -----
        `Result`:
            A Result object containing the result of the extraction process.
    """
    options = {
        "start_time": start_time,
        "stop_time": stop_time,
        "destdir": destdir,
        "filename": filename,
        "verbose": verbose,
        "overwrite": overwrite,
        "fps": fps,
        "dimensions": dimensions,
        "resize": resize,
        "rotate": rotate,
        "speed": speed,
        "bounce": bounce,
        "reverse": reverse,
        "monochrome": monochrome,
    }

    handler = ExtractionHandler(ExtractionMethod.GIF)
    return handler.execute(filepath, options)



import cv2  # type: ignore
import pytest

from videoxt.exceptions import ClosedVideoCaptureError
from videoxt.video import Video, fetch_video_properties, open_video_capture


def test_open_video_capture_if_filepath_as_path_and_is_valid_video_file(
    fixture_tmp_video_filepath,
):
    with open_video_capture(fixture_tmp_video_filepath) as opencap:
        assert opencap.isOpened()
        assert isinstance(opencap, cv2.VideoCapture)


def test_open_video_capture_if_filepath_as_str_and_is_valid_video_file(
    fixture_tmp_video_filepath,
):
    with open_video_capture(str(fixture_tmp_video_filepath)) as opencap:
        assert opencap.isOpened()
        assert isinstance(opencap, cv2.VideoCapture)


def test_open_video_capture_if_filepath_is_none():
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(None) as opencap:
            assert not opencap.isOpened()


def test_open_video_capture_if_filepath_is_int():
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(1) as opencap:
            assert not opencap.isOpened()


def test_open_video_capture_if_filepath_doesnt_exist(fixture_tmp_dir):
    non_existing_file = fixture_tmp_dir / "non_existing.mp4"
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(non_existing_file) as opencap:
            assert not opencap.isOpened()


def test_open_video_capture_if_filepath_is_a_directory(fixture_tmp_dir):
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(fixture_tmp_dir) as opencap:
            assert not opencap.isOpened()


def test_open_video_capture_if_filepath_is_an_existing_text_file(
    fixture_tmp_text_filepath,
):
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(fixture_tmp_text_filepath) as opencap:
            assert not opencap.isOpened()


def test_open_video_capture_if_filepath_is_a_video_file_with_zero_second_duration(
    fixture_tmp_video_filepath_zero_seconds,
):
    with pytest.raises(ClosedVideoCaptureError):
        with open_video_capture(fixture_tmp_video_filepath_zero_seconds) as opencap:
            assert not opencap.isOpened()


def test_fetch_video_properties_if_filepath_is_valid_video_file(
    fixture_tmp_video_filepath, fixture_tmp_video_properties
):
    fetched_properties = fetch_video_properties(fixture_tmp_video_filepath)
    assert isinstance(fetched_properties, dict)
    assert (
        fetched_properties["dimensions"] == fixture_tmp_video_properties["dimensions"]
    )
    assert fetched_properties["fps"] == fixture_tmp_video_properties["fps"]
    assert (
        fetched_properties["frame_count"] == fixture_tmp_video_properties["frame_count"]
    )


def test_video_object_is_instance_of_video_dataclass(fixture_tmp_video_filepath):
    video = Video(fixture_tmp_video_filepath)
    assert isinstance(video, Video)


def test_video_object_attributes_match_expected_values(
    fixture_tmp_video_filepath, fixture_tmp_video_properties
):
    video = Video(fixture_tmp_video_filepath)
    assert video.filepath == fixture_tmp_video_filepath
    assert video.dimensions == fixture_tmp_video_properties["dimensions"]
    assert video.fps == fixture_tmp_video_properties["fps"]
    assert video.frame_count == fixture_tmp_video_properties["frame_count"]
    assert video.duration == fixture_tmp_video_properties["duration"]
    assert video.duration_seconds == fixture_tmp_video_properties["duration_seconds"]
    assert (
        video.duration_timestamp == fixture_tmp_video_properties["duration_timestamp"]
    )
    assert video.has_audio == fixture_tmp_video_properties["has_audio"]
    assert (
        video.filesize_bytes
        == fixture_tmp_video_properties["video_file_path"].stat().st_size
    )


def test_video_object_with_invalid_filepath_raises_cv2_error(
    fixture_tmp_video_filepath_zero_seconds,
):
    with pytest.raises(ClosedVideoCaptureError):
        Video(fixture_tmp_video_filepath_zero_seconds)


def test_fetch_video_properties_with_invalid_filepath_raises_cv2_error(
    fixture_tmp_video_filepath_zero_seconds,
):
    with pytest.raises(ClosedVideoCaptureError):
        fetch_video_properties(fixture_tmp_video_filepath_zero_seconds)



"""Contains constants used throughout the libary and version information."""
import importlib.metadata
from enum import Enum

import cv2  # type: ignore

VERSION = importlib.metadata.version("videoxt")


class ExtractionMethod(Enum):
    """
    Enumeration of the supported extraction methods: audio, clip, frames, gif.

    Attributes:
    -----
        `AUDIO` (str):
            Extract the audio from a video file.
        `CLIP` (str):
            Extract a clip from a video file.
        `FRAMES` (str):
            Extract frames from a video file.
        `GIF` (str):
            Create a GIF from a video file.
    """

    AUDIO = "audio"
    CLIP = "clip"
    FRAMES = "frames"
    GIF = "gif"


SUPPORTED_VIDEO_FORMATS = {
    "3gp",
    "asf",
    "avi",
    "divx",
    "flv",
    "m4v",
    "mkv",
    "mov",
    "mp4",
    "mpeg",
    "mpg",
    "ogv",
    "rm",
    "ts",
    "vob",
    "webm",
    "wmv",
}

SUPPORTED_AUDIO_FORMATS = {
    "m4a",
    "mp3",
    "ogg",
    "wav",
}

SUPPORTED_IMAGE_FORMATS = {
    "bmp",
    "dib",
    "jp2",
    "jpeg",
    "jpg",
    "png",
    "tif",
    "tiff",
    "webp",
}

VALID_ROTATE_VALUES = {0, 90, 180, 270}

ROTATION_MAP = {
    90: cv2.ROTATE_90_CLOCKWISE,
    180: cv2.ROTATE_180,
    270: cv2.ROTATE_90_COUNTERCLOCKWISE,
}



"""Contains functions used to edit a VideoFileClip or np.ndarray prior to extraction."""
from typing import Optional

import cv2  # type: ignore
import numpy as np
from moviepy.editor import afx  # type: ignore
from moviepy.editor import VideoFileClip, vfx

import videoxt.constants as C


def trim_clip(
    clip: VideoFileClip,
    start_second: Optional[float] = None,
    stop_second: Optional[float] = None,
) -> VideoFileClip:
    """
    Trim a VideoFileClip to a specific time range and return the trimmed clip.

    Args:
    -----
        `clip` (moviepy.editor.VideoFileClip):
            The clip to trim.
        `start_second` (Optional[float]):
            The start time in seconds. If None, the trim will start at the beginning.
        `stop_second` (Optional[float]):
            The stop time in seconds. If None, the clip will be trimmed to the end.

    Returns:
    -----
        `moviepy.editor.VideoFileClip`: The trimmed clip.
    """
    if start_second is None and stop_second is None:
        return clip

    if start_second is None:
        start_second = 0.0

    if stop_second is None:
        stop_second = clip.duration

    return clip.subclip(start_second, stop_second)


def edit_clip_audio(
    clip: VideoFileClip,
    volume: Optional[float] = None,
    normalize: Optional[bool] = None,
) -> VideoFileClip:
    """
    Edit the audio of a VideoFileClip by adjusting its volume and normalizing the
    audio if specified and return the edited clip.

    Args:
    -----
        `clip` (moviepy.editor.VideoFileClip):
            The clip to edit.
        `volume` (Optional[float]):
            The volume multiplier. If None, the volume will not be adjusted.
        `normalize` (Optional[bool]):
            Whether to normalize the audio. If None, the audio will not be normalized.

    Returns:
    -----
        `moviepy.editor.VideoFileClip`: The edited clip.
    """
    if normalize:
        clip = clip.fx(afx.audio_normalize)

    if volume != 1.0:
        clip = clip.volumex(volume)

    return clip


def edit_clip_image(
    clip: VideoFileClip,
    dimensions: Optional[tuple[int, int]] = None,
    rotate: Optional[int] = None,
    monochrome: Optional[bool] = None,
) -> VideoFileClip:
    """
    Edit the image properties of a VideoFileClip by resizing, rotating, and converting
    to monochrome if specified and return the edited clip.

    Args:
    -----
        `clip` (moviepy.editor.VideoFileClip):
            The clip to edit.
        `dimensions` (Optional[tuple[int, int]]):
            The dimensions to resize the clip to. If None, the clip will not be resized.
        `rotate` (Optional[int]):
            The degrees to rotate the clip. If None, the clip will not be rotated.
        `monochrome` (Optional[bool]):
            Whether to convert the clip to monochrome. If None, the black and white
            filter will not be applied.

    Returns:
    -----
        `moviepy.editor.VideoFileClip`: The edited clip.
    """
    if dimensions is not None:
        clip = clip.resize(dimensions)

    if rotate != 0:
        clip = clip.rotate(rotate)

    if monochrome:
        clip = clip.fx(vfx.blackwhite)

    return clip


def edit_clip_motion(
    clip: VideoFileClip,
    speed: Optional[float] = None,
    reverse: Optional[bool] = None,
    bounce: Optional[bool] = None,
) -> VideoFileClip:
    """
    Edit the moving image properties of a VideoFileClip by adjusting its speed,
    reversing, and bouncing if specified and return the edited clip.

    Args:
    -----
        `clip` (moviepy.editor.VideoFileClip):
            The clip to edit.
        `speed` (Optional[float]):
            The speed multiplier. If None, the speed will not be adjusted.
        `reverse` (Optional[bool]):
            Whether to reverse the clip. If None, the clip will not be reversed.
        `bounce` (Optional[bool]):
            Whether to bounce the clip. If None, the clip will not be bounced.

    Returns:
    -----
        `moviepy.editor.VideoFileClip`: The edited clip.
    """
    if reverse:
        clip = clip.fx(vfx.time_mirror)

    if bounce:
        clip = clip.fx(vfx.time_symmetrize)

    if speed != 1.0:
        clip = clip.fx(vfx.speedx, speed)

    return clip


def edit_image(
    image: np.ndarray,
    dimensions: Optional[tuple[int, int]] = None,
    rotate: Optional[int] = None,
    monochrome: Optional[bool] = None,
) -> np.ndarray:
    """
    Edit a numpy.ndarray image by resizing, rotating, and converting to monochrome
    if specified and return the edited image.

    Args:
    -----
        `image` (np.ndarray):
            The image to edit.
        `dimensions` (Optional[tuple[int, int]]):
            The dimensions to resize the image to. If None, the image will not be
            resized.
        `rotate` (Optional[int]):
            The degrees to rotate the image. If None, the image will not be rotated.
        `monochrome` (Optional[bool]):
            Whether to convert the image to monochrome. If None, the black and white
            filter will not be applied.

    Returns:
    -----
        `np.ndarray`: The edited image.
    """
    if dimensions is not None and dimensions != image.shape:
        image = cv2.resize(image, dimensions)

    if rotate != 0 and rotate is not None:
        try:
            rotate_value = C.ROTATION_MAP[rotate]
        except KeyError:
            pass  # XXX: log
        else:
            image = cv2.rotate(image, rotate_value)

    if monochrome:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    return image



"""
Command line interface for the videoxt library.

Usage:
-----
$ videoxt --help
usage: videoxt [-h] [--version] {audio,clip,frames,gif} ...

Extract audio, individual frames, short clips and GIFs from videos.

positional arguments:
  {audio,clip,frames,gif}
    audio               Extract audio from a video file.
    clip                Extract a short clip from a video file as `mp4`.
    frames              Extract individual frames from a video and save them as images.
    gif                 Create a GIF from a video between two specified points.

options:
  -h, --help            show this help message and exit
  --version, -V         show program's version number and exit
"""
import argparse
from collections.abc import Sequence
from functools import partial
from typing import Any, Optional

import videoxt.api
import videoxt.validators as V
from videoxt.constants import (
    SUPPORTED_AUDIO_FORMATS,
    SUPPORTED_IMAGE_FORMATS,
    VALID_ROTATE_VALUES,
    VERSION,
)
from videoxt.exceptions import VideoXTError


def split_cli_args(args: argparse.Namespace) -> tuple[str, str, dict[str, Any]]:
    """
    Split the arguments into a tuple containing the extraction method, filepath, and
    extraction options. The extraction method is determined by the subparser name,
    which can be one of either "audio", "clip", "frames", or "gif".

    Args:
    -----
        `args` (argparse.Namespace): The arguments from the CLI.

    Returns:
    -----
        `tuple[str, str, dict[str, Any]]`:
            The extraction method, filepath, and extraction options.
    """
    options = {
        k: v for k, v in vars(args).items() if k not in ["subparser_name", "filepath"]
    }

    return args.subparser_name, args.filepath, options


def execute_extraction(
    method: str,
    filepath: str,
    **options: dict,
) -> int:
    """
    Trigger the extraction procedure and return the exit code (0 means success).

    Args:
    -----
        `method` (str]):
            The extraction method to use ("audio", "clip", "frames", "gif").
        `filepath` (str):
            Path to the video file.
        `**options` (dict):
            Extraction options specific to the chosen extraction method.

    Returns:
    -----
        `int` : 0 if the extraction was successful, 1 otherwise.
    """

    try:
        videoxt.api.extract(method, filepath, skip_validation=True, **options)
    except VideoXTError as error_msg:
        print(error_msg)
        return 1
    else:
        return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    """
    The main entry point when called from the command-line. By default, `verbose` mode
    is enabled, which prints details about the prepared extraction request and the
    result. To disable this output, use the `--quiet` or `-q` flag.

    Args:
    -----
        `argv` (Optional[Sequence[str]]): The arguments from the CLI.

    Returns:
    -----
        `int`: 0 if the extraction was successful, 1 otherwise.

    Raises:
    -----
        `argparse.ArgumentTypeError`:
            Raised when an argument is not of the expected type.
    """
    # parent_parser houses arguments common to all subparsers
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "filepath",
        type=partial(V.valid_filepath, is_video=True),
        help="Path to the video file with extension.",
    )
    parent_parser.add_argument(
        "--start-time",
        "-s",
        type=V.valid_start_time,
        default=0,
        metavar="",
        dest="start_time",
        help=(
            "Time to start extraction. Can be a number representing seconds or a "
            "timestamp (Ex: --start-time 0:45 or -s 45)."
        ),
    )
    parent_parser.add_argument(
        "--stop-time",
        "-S",
        type=V.valid_stop_time,
        metavar="",
        dest="stop_time",
        help=(
            "Time to stop extraction. Can be a number representing seconds or a "
            "timestamp (Ex: --stop-time 1:30 or -S 90)."
        ),
    )
    parent_parser.add_argument(
        "--destdir",
        "-d",
        type=V.valid_dir,
        metavar="",
        dest="destdir",
        help=(
            "Specify the directory you want to save output to. If not provided, "
            "media is saved in the directory of the input video file."
        ),
    )
    parent_parser.add_argument(
        "--filename",
        "-fn",
        type=V.valid_filename,
        metavar="",
        dest="filename",
        help=(
            "Set the name of the output media file(s), without the extension. "
            "If not provided, the video's filename is used."
        ),
    )
    parent_parser.add_argument(
        "--quiet",
        "-q",
        action="store_false",
        dest="verbose",
        help="Disable extraction details from being printed to the console.",
    )
    parent_parser.add_argument(
        "--overwrite",
        "-ov",
        action="store_true",
        dest="overwrite",
        help="Overwrite the output file(s) if they already exist.",
    )
    parent_parser.add_argument(
        "--fps",
        "-f",
        type=V.valid_fps,
        metavar="",
        help=(
            "Manually set the video's frames per second (FPS). "
            "Helpful if the FPS is not read accurately by OpenCV. Use with caution."
        ),
    )

    # parent_parser_audio houses arguments common to audio and clip subparsers
    parent_parser_audio = argparse.ArgumentParser(add_help=False)
    parent_parser_audio.add_argument(
        "--volume",
        "-v",
        type=V.valid_volume,
        default=1.0,
        metavar="",
        help="Increase or decrease the output audio volume by a factor of N.",
    )
    parent_parser_audio.add_argument(
        "--normalize",
        action="store_true",
        help="Normalize the audio output to a maximum of 0dB.",
    )

    # parent_parser_image houses arguments common to clip, frames and gif subparsers
    parent_parser_image = argparse.ArgumentParser(add_help=False)
    parent_parser_image.add_argument(
        "--dimensions",
        "-dm",
        type=V.valid_dimensions_str,
        metavar="",
        help="Resize the output to a specific width and height (Ex: -dm 1920x1080).",
    )
    parent_parser_image.add_argument(
        "--resize",
        "-rs",
        type=V.valid_resize,
        default=1.0,
        metavar="",
        help="Increase or decrease the dimensions of the output by a factor of N.",
    )
    parent_parser_image.add_argument(
        "--rotate",
        "-rt",
        type=V.valid_rotate_value,
        choices=VALID_ROTATE_VALUES,
        default=0,
        metavar="",
        dest="rotate",
        help="Rotate the output by 90, 180, or 270 degrees.",
    )
    parent_parser_image.add_argument(
        "--monochrome",
        action="store_true",
        help="Apply a black-and-white filter to the output.",
    )

    # parent_parser_motion houses arguments common to audio, clip and gif subparsers
    parent_parser_motion = argparse.ArgumentParser(add_help=False)
    parent_parser_motion.add_argument(
        "--speed",
        "-sp",
        type=V.valid_speed,
        default=1.0,
        metavar="",
        help="Increase or decrease the speed of the output by a factor of N.",
    )
    parent_parser_motion.add_argument(
        "--bounce",
        action="store_true",
        help="Make the output bounce back-and-forth, boomerang style.",
    )
    parent_parser_motion.add_argument(
        "--reverse",
        action="store_true",
        help="Reverse the output.",
    )

    # main_parser encompasses all subparsers
    main_parser = argparse.ArgumentParser(
        prog="videoxt",
        description=(
            "Extract audio, individual frames, short clips and GIFs from videos."
        ),
    )
    main_parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {VERSION}",
    )
    subparsers = main_parser.add_subparsers(dest="subparser_name", required=True)

    # audio subparser
    subparser_audio = subparsers.add_parser(
        "audio",
        help="Extract audio from a video file.",
        parents=[parent_parser, parent_parser_audio, parent_parser_motion],
    )
    subparser_audio.add_argument(
        "--audio-format",
        "-af",
        type=V.valid_audio_format,
        choices=SUPPORTED_AUDIO_FORMATS,
        default="mp3",
        metavar="",
        dest="audio_format",
        help="Set the extracted audio file format. Default is 'mp3'.",
    )

    # clip subparser
    subparsers.add_parser(
        "clip",
        help="Extract a short clip from a video file as 'mp4'.",
        parents=[
            parent_parser,
            parent_parser_audio,
            parent_parser_image,
            parent_parser_motion,
        ],
    )

    # frames subparser
    subparser_frames = subparsers.add_parser(
        "frames",
        help="Extract individual frames from a video and save them as images.",
        parents=[parent_parser, parent_parser_image],
    )
    subparser_frames.add_argument(
        "--image-format",
        "-if",
        type=V.valid_image_format,
        choices=SUPPORTED_IMAGE_FORMATS,
        default="jpg",
        metavar="",
        dest="image_format",
        help="Set the image format to save the frames as. Default is 'jpg'.",
    )
    subparser_frames.add_argument(
        "--capture-rate",
        "-cr",
        type=V.valid_capture_rate,
        default=1,
        metavar="",
        dest="capture_rate",
        help="Capture every Nth video frame. Default is 1, which captures every frame.",
    )

    # gif subparser
    subparsers.add_parser(
        "gif",
        help="Create a GIF from a video between two specified points.",
        parents=[parent_parser, parent_parser_image, parent_parser_motion],
    )

    # Parse the arguments and execute the extraction.
    try:
        args = main_parser.parse_args(argv)
    except (argparse.ArgumentTypeError, VideoXTError) as error_msg:
        print(error_msg)
        return 1
    else:
        method, filepath, options = split_cli_args(args)
        return execute_extraction(method, filepath, **options)


if __name__ == "__main__":
    exit(main())



"""Contains Request models that validate and prepare a request for extraction."""
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import videoxt.preppers as P
import videoxt.validators as V
from videoxt.exceptions import NoAudioError
from videoxt.utils import ToJsonMixin
from videoxt.video import Video


@dataclass
class Request(ABC):
    """
    Protocol defining the interface for validating and preparing extraction requests.

    Public Methods:
    -----
        `validate()` -> `Request`:
            Validate fields if specified and return a validated request of the same
            type.
        `prepare()` -> `PreparedRequest`:
            Construct and return a prepared request using a video and a validated
            request of some type. A prepared request is required by an extractor of the
            same type.

    Properties:
    -----
        `is_validated` (bool): If the request has been validated.
    """

    @abstractmethod
    def validate(self) -> "Request":
        """
        Validate fields if specified and return a validated request of the same type.
        """
        ...

    @abstractmethod
    def prepare(self, video: Video) -> "PreparedRequest":
        """
        Construct and return a prepared request using a video and a validated request
        of some type. A prepared request is required by an extractor of the same type.
        """
        ...

    @property
    @abstractmethod
    def is_validated(self) -> bool:
        """Return True if `Request` has been validated, False otherwise."""
        ...


@dataclass
class BaseRequest(Request):
    """
    Stores, validates and prepares extraction request arguments shared by all methods.

    Fields
    -----
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.

    Public Methods
    -----
        - `validate()` -> `BaseRequest`:
            Validate fields if specified and return a validated `BaseRequest`.
        - `prepare()` -> `PreparedBaseRequest`:
            Construct and return a prepared base request using a video and a base
            request. A prepared base request is not used by any extractor. Its sole
            purpose is to validate and prepare the shared request arguments.

    Properties
    -----
        - `is_validated` -> `bool`: If the request has been validated.
    """

    start_time: Optional[float | int | str] = None  # XXX: add timedelta support
    stop_time: Optional[float | int | str] = None  # XXX: add timedelta support
    destdir: Optional[Path] = None
    filename: Optional[str] = None
    verbose: Optional[bool] = None
    overwrite: Optional[bool] = None
    fps: Optional[float] = None
    _is_validated: bool = field(init=False)

    def validate(self) -> "BaseRequest":
        """Validate fields if specified and return a validated `BaseRequest`."""
        self._validate_start_time()
        self._validate_stop_time()
        self._validate_fps()
        self._validate_destdir()
        self._validate_filename()
        return self

    def prepare(self, video: Video) -> "PreparedBaseRequest":
        """
        Construct and return a `PreparedBaseRequest` using self and a `Video`.

        Args:
        -----
            `video` (Video): Contains video metadata used to prepare request.

        Returns:
        -----
            `PreparedBaseRequest`: A prepared base extraction request.
        """
        p = PreparedBaseRequest(
            video=video,
            start_time=self.start_time,
            stop_time=self.stop_time,
            fps=self.fps,
            destdir=self.destdir,
            filename=self.filename,
            verbose=self.verbose,
        )
        p.prepare()
        return p

    @property
    def is_validated(self) -> bool:
        """Return True if `Request` has been validated, False otherwise."""
        return self._is_validated

    def _validate_start_time(self) -> float | int | str | None:
        """Validate the requested start time if specified."""
        self.start_time = (
            None if self.start_time is None else V.valid_start_time(self.start_time)
        )
        return self.start_time

    def _validate_stop_time(self) -> float | int | str | None:
        """Validate the requested stop time if specified."""
        self.stop_time = (
            None if self.stop_time is None else V.valid_stop_time(self.stop_time)
        )
        return self.stop_time

    def _validate_fps(self) -> float | None:
        """Validate the requested fps if specified."""
        self.fps = None if self.fps is None else V.positive_float(self.fps)
        return self.fps

    def _validate_destdir(self) -> Path | None:
        """Validate the requested destination directory if specified."""
        self.destdir = None if self.destdir is None else V.valid_dir(self.destdir)
        return self.destdir

    def _validate_filename(self) -> str | None:
        """Validate the requested filename if specified."""
        self.filename = (
            None if self.filename is None else V.valid_filename(self.filename)
        )
        return self.filename


@dataclass
class AudioRequest(BaseRequest):
    """
    Stores, validates and prepares 'audio' extraction request arguments.

    Fields
    -----
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `audio_format` (Optional[str]):
            Set the extracted audio file format. Defaults to 'mp3' if not specified.
            See: `videoxt.constants.SUPPORTED_AUDIO_FORMATS`.
        `speed` (Optional[float]):
            Set the speed of the extracted audio. A value of 0.5 will halve the speed of
            the extracted audio. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted audio bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted audio. Defaults to False if not specified.
        `volume` (Optional[float]):
            Set the volume of the extracted audio. A value of 0.5 will halve the volume
            of the extracted audio. Defaults to 1.0 if not specified (no change).
        `normalize` (Optional[bool]):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Public Methods
    -----
        - `validate()` -> `AudioRequest`:
            Validate fields if specified and return a validated `AudioRequest`.
        - `prepare()` -> `PreparedAudioRequest`:
            Construct and return a prepared audio request using a video and a validated
            `AudioRequest`. A prepared audio request is required by an `AudioExtractor`.

    Properties
    -----
        - `is_validated` -> `bool`: If the request has been validated.
    """

    audio_format: Optional[str] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    volume: Optional[float] = None
    normalize: Optional[bool] = None

    def validate(self) -> "AudioRequest":
        """Validate fields if specified and return a validated `AudioRequest`."""
        super().validate()
        self._validate_audio_format()
        self._validate_speed()
        self._validate_volume()
        self._is_validated = True
        return self

    def prepare(self, video: Video) -> "PreparedAudioRequest":
        """
        Construct and return a `PreparedAudioRequest` using self and a `Video`.

        Fields will validate before preparation if not validated.

        Args:
        -----
            `video` (Video): Contains video metadata used to prepare request.

        Returns:
        -----
            `PreparedAudioRequest`:
                A prepared audio extraction request, required by an `AudioExtractor`.
        """
        if not self.is_validated:
            self.validate()  # XXX: log

        p = PreparedAudioRequest(
            video=video,
            start_time=self.start_time,
            stop_time=self.stop_time,
            destdir=self.destdir,
            filename=self.filename,
            verbose=self.verbose,
            overwrite=self.overwrite,
            fps=self.fps,
            audio_format=self.audio_format,
            speed=self.speed,
            bounce=self.bounce,
            reverse=self.reverse,
            volume=self.volume,
            normalize=self.normalize,
        )
        p.prepare()
        return p

    def _validate_audio_format(self) -> str | None:
        """Validate the requested audio format if specified."""
        self.audio_format = (
            None
            if self.audio_format is None
            else V.valid_audio_format(self.audio_format)
        )
        return self.audio_format

    def _validate_speed(self) -> float | None:
        """Validate the requested speed if specified."""
        self.speed = None if self.speed is None else V.positive_float(self.speed)
        return self.speed

    def _validate_volume(self) -> float | None:
        """Validate the requested volume if specified."""
        self.volume = None if self.volume is None else V.valid_volume(self.volume)
        return self.volume


@dataclass
class ClipRequest(BaseRequest):
    """
    Stores, validates and prepares 'clip' extraction request arguments.

    Recommended usage: Set a short extraction range. The process can be slow for long
    or high-resolution videos.

    Fields
    -----
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the clip. Defaults to
            the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the clip by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the clip by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (Optional[float]):
            Set the speed of the extracted clip. A value of 0.5 will halve the playback
            speed of the clip. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted clip bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted clip. Defaults to False if not specified.
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the clip. Defaults to False if
            not specified.
        `volume` (Optional[float]):
            Set the volume of the extracted clip's audio. A value of 0.5 will halve the
            volume of the clip's audio. Defaults to 1.0 if not specified (no change).
        `normalize` (Optional[bool]):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Public Methods
    -----
        - `validate()` -> `ClipRequest`:
            Validate fields if specified and return a validated `ClipRequest`.
        - `prepare()` -> `PreparedClipRequest`:
            Construct and return a prepared clip request using a video and a validated
            `ClipRequest`. A prepared clip request is required by a `ClipExtractor`.

    Properties
    -----
        - `is_validated` -> `bool`: If the request has been validated.
    """

    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    monochrome: Optional[bool] = None
    volume: Optional[float] = None
    normalize: Optional[bool] = None

    def validate(self) -> "ClipRequest":
        """Validate fields if specified and return a validated `ClipRequest`."""
        super().validate()
        self._validate_resize()
        self._validate_dimensions()
        self._validate_rotate()
        self._validate_speed()
        self._validate_volume()
        self._is_validated = True
        return self

    def prepare(self, video: Video) -> "PreparedClipRequest":
        """
        Construct and return a `PreparedClipRequest` using self and a `Video`.

        Fields will validate before preparation if not validated.

        Args:
        -----
            `video` (Video): Contains video metadata used to prepare request.

        Returns:
        -----
            `PreparedClipRequest`:
                A prepared clip extraction request, required by a `ClipExtractor`.
        """
        if not self.is_validated:
            self.validate()  # XXX: log

        p = PreparedClipRequest(
            video=video,
            start_time=self.start_time,
            stop_time=self.stop_time,
            destdir=self.destdir,
            filename=self.filename,
            verbose=self.verbose,
            overwrite=self.overwrite,
            fps=self.fps,
            dimensions=self.dimensions,
            resize=self.resize,
            rotate=self.rotate,
            speed=self.speed,
            bounce=self.bounce,
            reverse=self.reverse,
            monochrome=self.monochrome,
            volume=self.volume,
            normalize=self.normalize,
        )
        p.prepare()
        return p

    def _validate_resize(self) -> float | None:
        """Validate the requested resize value if specified."""
        self.resize = None if self.resize is None else V.positive_float(self.resize)
        return self.resize

    def _validate_dimensions(self) -> tuple[int, int] | None:
        """Validate the requested dimensions if specified."""
        self.dimensions = (
            None if self.dimensions is None else V.valid_dimensions(self.dimensions)
        )
        return self.dimensions

    def _validate_rotate(self) -> int | None:
        """Validate the requested rotation if specified."""
        self.rotate = None if self.rotate is None else V.valid_rotate_value(self.rotate)
        return self.rotate

    def _validate_speed(self) -> float | None:
        """Validate the requested speed if specified."""
        self.speed = None if self.speed is None else V.positive_float(self.speed)
        return self.speed

    def _validate_volume(self) -> float | None:
        """Validate the requested volume if specified."""
        self.volume = None if self.volume is None else V.valid_volume(self.volume)
        return self.volume


@dataclass
class FramesRequest(BaseRequest):
    """
    Stores, validates and prepares 'frames' extraction request arguments.

    The images are saved to a directory named after the video file, or to a directory
    you specify.

    Fields
    -----
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `image_format` (Optional[str]):
            Set the extracted image file format. Defaults to 'jpg' if not specified.
            See: `videoxt.constants.SUPPORTED_IMAGE_FORMATS`.
        `capture_rate` (Optional[int]):
            Capture every Nth video frame. Defaults to 1 if not specified, which
            extracts every frame within the extraction range.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the images. Defaults
            to the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the images by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the images by `n` degrees. Allowed values: 0, 90, 180 or 270.
            Defaults to 0 if not specified (no change).
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the images. Defaults to False if
            not specified.

    Public Methods
    -----
        - `validate()` -> `FramesRequest`:
            Validate fields if specified and return a validated `FramesRequest`.
        - `prepare()` -> `PreparedFramesRequest`:
            Construct and return a prepared frames request using a video and a validated
            `FramesRequest`. A prepared frames request is required by a
            `FramesExtractor`.

    Properties
    -----
        - `is_validated` -> `bool`: If the request has been validated.
    """

    image_format: Optional[str] = None
    capture_rate: Optional[int] = None
    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    monochrome: Optional[bool] = None

    def validate(self) -> "FramesRequest":
        """Validate fields if specified and return a validated `FramesRequest`."""
        super().validate()
        self._validate_image_format()
        self._validate_capture_rate()
        self._validate_resize()
        self._validate_dimensions()
        self._validate_rotate()
        self._is_validated = True
        return self

    def prepare(self, video: Video) -> "PreparedFramesRequest":
        """
        Construct and return a `PreparedFramesRequest` using self and a `Video`.

        Fields will validate before preparation if not validated.

        Args:
        -----
            `video` (Video): Contains video metadata used to prepare request.

        Returns:
        -----
            `PreparedFramesRequest`:
                A prepared frames extraction request, required by a `FramesExtractor`.
        """
        if not self.is_validated:
            self.validate()  # XXX: log

        p = PreparedFramesRequest(
            video=video,
            start_time=self.start_time,
            stop_time=self.stop_time,
            destdir=self.destdir,
            filename=self.filename,
            verbose=self.verbose,
            overwrite=self.overwrite,
            fps=self.fps,
            image_format=self.image_format,
            capture_rate=self.capture_rate,
            dimensions=self.dimensions,
            resize=self.resize,
            rotate=self.rotate,
            monochrome=self.monochrome,
        )
        p.prepare()
        return p

    def _validate_image_format(self) -> str | None:
        """Validate the requested image format if specified."""
        self.image_format = (
            None
            if self.image_format is None
            else V.valid_image_format(self.image_format)
        )
        return self.image_format

    def _validate_capture_rate(self) -> int | None:
        """Validate the requested capture rate if specified."""
        self.capture_rate = (
            None if self.capture_rate is None else V.positive_int(self.capture_rate)
        )

        return self.capture_rate

    def _validate_resize(self) -> float | None:
        """Validate the requested resize value if specified."""
        self.resize = None if self.resize is None else V.positive_float(self.resize)
        return self.resize

    def _validate_dimensions(self) -> tuple[int, int] | None:
        """Validate the requested dimensions if specified."""
        self.dimensions = (
            None if self.dimensions is None else V.valid_dimensions(self.dimensions)
        )
        return self.dimensions

    def _validate_rotate(self) -> int | None:
        """Validate the requested rotation if specified."""
        self.rotate = None if self.rotate is None else V.valid_rotate_value(self.rotate)
        return self.rotate


@dataclass
class GifRequest(BaseRequest):
    """
    Stores, validates and prepares 'gif' extraction request arguments.

    Recommended usage: Set a short extraction range. The process can be slow for long
    or high-resolution videos.

    Fields
    -----
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the gif. Defaults to
            the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the gif by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the gif by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (Optional[float]):
            Set the speed of the extracted gif. A value of 0.5 will halve the playback
            speed of the gif. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted gif bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted gif. Defaults to False if not specified.
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the gif. Defaults to False if
            not specified.

    Public Methods
    -----
        - `validate()` -> `GifRequest`:
            Validate fields if specified and return a validated `GifRequest`.
        - `prepare()` -> `PreparedGifRequest`:
            Construct and return a prepared gif request using a video and a validated
            `GifRequest`. A prepared gif request is required by a `GifExtractor`.

    Properties
    -----
        - `is_validated` -> `bool`: If the request has been validated.
    """

    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    monochrome: Optional[bool] = None

    def validate(self) -> "GifRequest":
        """Validate fields if specified and return a validated `GifRequest`."""
        super().validate()
        self._validate_resize()
        self._validate_dimensions()
        self._validate_rotate()
        self._validate_speed()
        self._is_validated = True
        return self

    def prepare(self, video: Video) -> "PreparedGifRequest":
        """
        Construct and return a `PreparedGifRequest` using self and a `Video`.

        Fields will validate before preparation if not validated.

        Args:
        -----
            `video` (Video): Contains video metadata used to prepare request.

        Returns:
        -----
            `PreparedGifRequest`:
                A prepared gif extraction request, required by a `GifExtractor`.
        """
        if not self.is_validated:
            self.validate()  # XXX: log

        p = PreparedGifRequest(
            video=video,
            start_time=self.start_time,
            stop_time=self.stop_time,
            destdir=self.destdir,
            filename=self.filename,
            verbose=self.verbose,
            overwrite=self.overwrite,
            fps=self.fps,
            dimensions=self.dimensions,
            resize=self.resize,
            rotate=self.rotate,
            speed=self.speed,
            bounce=self.bounce,
            reverse=self.reverse,
            monochrome=self.monochrome,
        )
        p.prepare()
        return p

    def _validate_resize(self) -> float | None:
        """Validate the requested resize value if specified."""
        self.resize = None if self.resize is None else V.positive_float(self.resize)
        return self.resize

    def _validate_dimensions(self) -> tuple[int, int] | None:
        """Validate the requested dimensions if specified."""
        self.dimensions = (
            None if self.dimensions is None else V.valid_dimensions(self.dimensions)
        )
        return self.dimensions

    def _validate_rotate(self) -> int | None:
        """Validate the requested rotation if specified."""
        self.rotate = None if self.rotate is None else V.valid_rotate_value(self.rotate)
        return self.rotate

    def _validate_speed(self) -> float | None:
        """Validate the requested speed if specified."""
        self.speed = None if self.speed is None else V.positive_float(self.speed)
        return self.speed


@dataclass
class PreparedRequest(ABC, ToJsonMixin):
    """
    Protocol defining the interface for preparing requests for extraction.

    Public Methods:
    -----
        - `prepare()` -> `PreparedRequest`:
            Prepare fields and return a fully prepared request of some type.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared request type.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    @abstractmethod
    def prepare(self) -> "PreparedRequest":
        """Prepare fields and return a fully prepared request of some type."""
        ...

    @property
    @abstractmethod
    def is_prepared(self) -> bool:
        """Return True if the request has been prepared, False otherwise."""
        ...

    @property
    @abstractmethod
    def is_verbose(self) -> bool | None:
        """Return the state of the verbose field."""
        ...


@dataclass
class PreparedBaseRequest(PreparedRequest):
    """
    Prepares and stores request arguments shared by all extraction methods.

    Fields
    -----
        `video` (Video):
            The object containing video metadata used to prepare the request.
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `destpath` (pathlib.Path):
            The destination path of the extracted file or directory. Not initialized.
        `extraction_range` (dict):
            The range represented as seconds, as timestamps, and frame numbers.
            Not initialized.

    Public Methods
    -----
        - `prepare()` -> `PreparedBaseRequest`:
            Prepare fields and return a fully prepared base request.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared base request.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    video: Video
    start_time: Optional[float | int | str] = None
    stop_time: Optional[float | int | str] = None
    destdir: Optional[Path] = None
    filename: Optional[str] = None
    verbose: Optional[bool] = None
    overwrite: Optional[bool] = None
    fps: Optional[float] = None
    destpath: Path = field(init=False)
    extraction_range: dict = field(init=False, default_factory=dict)
    _is_prepared: bool = field(init=False)

    def prepare(self) -> "PreparedBaseRequest":
        """Prepare fields and return a `PreparedBaseRequest`."""
        self._prepare_start_time()
        self._prepare_stop_time()
        self._prepare_fps()
        self._prepare_extraction_range()
        self._prepare_verbose()
        self._prepare_overwrite()
        return self

    @property
    def is_prepared(self) -> bool:
        """Return True if the request has been prepared, False otherwise."""
        return self._is_prepared

    @property
    def is_verbose(self) -> bool | None:
        """Return the state of the verbose field."""
        return self.verbose

    def _prepare_start_time(self) -> float | int | str:
        """Set start time to 0 if not specified."""
        self.start_time = P.prepare_start_time(self.start_time)
        return self.start_time

    def _prepare_stop_time(self) -> float | int | str:
        """Set stop time to video duration if not specified."""
        self.stop_time = P.prepare_stop_time(
            self.video.duration_seconds, self.stop_time
        )
        return self.stop_time

    def _prepare_fps(self) -> float:
        """Set fps to video fps if not specified."""
        self.fps = P.prepare_fps(self.video.fps, self.fps)
        return self.fps

    def _prepare_extraction_range(self) -> dict:
        """Prepare, set and return a dictionary representing the extraction range."""
        self.extraction_range = P.prepare_extraction_range(
            self.video.duration_seconds,
            self.video.frame_count,
            self.start_time,
            self.stop_time,
            self.fps,
        )

        return self.extraction_range

    def _prepare_verbose(self) -> bool:
        """Set verbose to False if not specified."""
        if self.verbose is None:
            self.verbose = False
        return self.verbose

    def _prepare_overwrite(self) -> bool:
        """Set overwrite to False if not specified."""
        if self.overwrite is None:
            self.overwrite = False
        return self.overwrite


@dataclass
class PreparedAudioRequest(PreparedBaseRequest):
    """
    Prepares and stores 'audio' extraction request arguments.

    Fields
    -----
        `video` (Video):
            The object containing video metadata used to prepare the request.
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `destpath` (pathlib.Path):
            The destination path of the extracted file or directory. Not initialized.
        `extraction_range` (dict):
            The range represented as seconds, as timestamps, and frame numbers.
            Not initialized.
        `audio_format` (Optional[str]):
            Set the extracted audio file format. Defaults to 'mp3' if not specified.
            See: `videoxt.constants.SUPPORTED_AUDIO_FORMATS`.
        `speed` (Optional[float]):
            Set the speed of the extracted audio. A value of 0.5 will halve the speed of
            the extracted audio. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted audio bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted audio. Defaults to False if not specified.
        `volume` (Optional[float]):
            Set the volume of the extracted audio. A value of 0.5 will halve the volume
            of the extracted audio. Defaults to 1.0 if not specified (no change).
        `normalize` (Optional[bool]):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Public Methods
    -----
        - `prepare()` -> `PreparedAudioRequest`:
            Prepare fields and return a fully prepared audio request, required by an
            `AudioExtractor`.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared audio request.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    audio_format: Optional[str] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    volume: Optional[float] = None
    normalize: Optional[bool] = None

    def prepare(self) -> "PreparedAudioRequest":
        """
        Prepare request fields and return a `PreparedAudioRequest`, required by an
        `AudioExtractor`.
        """
        self._check_video_has_audio()
        super().prepare()
        self._prepare_audio_format()
        self._prepare_destpath()
        self._prepare_speed()
        self._prepare_bounce()
        self._prepare_reverse()
        self._prepare_volume()
        self._prepare_normalize()
        self._is_prepared = True
        return self

    def _check_video_has_audio(self) -> None:
        """Raise `NoAudioError` if the video does not have audio."""
        if not self.video.has_audio:
            raise NoAudioError("Video does not have audio.")

    def _prepare_audio_format(self) -> str:
        """Set audio format to 'mp3' if not specified."""
        if self.audio_format is None:
            self.audio_format = "mp3"
        return self.audio_format

    def _prepare_destpath(self) -> Path:
        """Set the destination path the extracted audio file will be saved to."""
        self.destpath = P.prepare_destpath(
            self.video.filepath,
            self.filename,
            self.destdir,
            self.audio_format,
            self.overwrite,
        )

        return self.destpath

    def _prepare_speed(self) -> float:
        """Set speed to 1.0 if not specified."""
        if self.speed is None:
            self.speed = 1.0
        return self.speed

    def _prepare_bounce(self) -> bool:
        """Set bounce to False if not specified."""
        if self.bounce is None:
            self.bounce = False
        return self.bounce

    def _prepare_reverse(self) -> bool:
        """Set reverse to False if not specified."""
        if self.reverse is None:
            self.reverse = False
        return self.reverse

    def _prepare_volume(self) -> float:
        """Set volume to 1.0 if not specified."""
        if self.volume is None:
            self.volume = 1.0
        return self.volume

    def _prepare_normalize(self) -> bool:
        """Set normalize to False if not specified."""
        if self.normalize is None:
            self.normalize = False
        return self.normalize


@dataclass
class PreparedClipRequest(PreparedBaseRequest):
    """
    Prepares and stores 'clip' extraction request arguments.

    Fields
    -----
        `video` (Video):
            The object containing video metadata used to prepare the request.
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `destpath` (pathlib.Path):
            The destination path of the extracted file or directory. Not initialized.
        `extraction_range` (dict):
            The range represented as seconds, as timestamps, and frame numbers.
            Not initialized.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the clip. Defaults to
            the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the clip by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the clip by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (Optional[float]):
            Set the speed of the extracted clip. A value of 0.5 will halve the playback
            speed of the clip. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted clip bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted clip. Defaults to False if not specified.
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the clip. Defaults to False if
            not specified.
        `volume` (Optional[float]):
            Set the volume of the extracted clip's audio. A value of 0.5 will halve the
            volume of the clip's audio. Defaults to 1.0 if not specified (no change).
        `normalize` (Optional[bool]):
            If True, normalize the audio. Normalization adjusts the gain of the audio to
            ensure consistent levels, preventing distortion and enhancing clarity in
            some cases. Defaults to False if not specified.

    Public Methods
    -----
        - `prepare()` -> `PreparedClipRequest`:
            Prepare fields and return a fully prepared clip request, required by a
            `ClipExtractor`.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared clip request.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    monochrome: Optional[bool] = None
    volume: Optional[float] = None
    normalize: Optional[bool] = None

    def prepare(self) -> "PreparedClipRequest":
        """
        Prepare request fields and return a `PreparedClipRequest`, required by a
        `ClipExtractor`.
        """
        super().prepare()
        self._prepare_destpath()
        self._prepare_resize()
        self._prepare_dimensions()
        self._prepare_rotate()
        self._prepare_speed()
        self._prepare_bounce()
        self._prepare_reverse()
        self._prepare_monochrome()
        self._prepare_volume()
        self._prepare_normalize()
        self._is_prepared = True
        return self

    def _prepare_destpath(self) -> Path:
        """Set the destination path the extracted clip will be saved to."""
        self.destpath = P.prepare_destpath(
            self.video.filepath,
            self.filename,
            self.destdir,
            ".mp4",
            self.overwrite,
        )

        return self.destpath

    def _prepare_resize(self) -> float:
        """Set resize to 1.0 if not specified."""
        if self.resize is None:
            self.resize = 1.0
        return self.resize

    def _prepare_dimensions(self) -> tuple[int, int]:
        """Set dimensions to the video dimensions if not specified."""
        self.dimensions = P.prepare_dimensions(
            self.video.dimensions, self.dimensions, self.resize
        )
        return self.dimensions

    def _prepare_rotate(self) -> int:
        """Set rotate to 0 if not specified."""
        if self.rotate is None:
            self.rotate = 0
        return self.rotate

    def _prepare_speed(self) -> float:
        """Set speed to 1.0 if not specified."""
        if self.speed is None:
            self.speed = 1.0
        return self.speed

    def _prepare_bounce(self) -> bool:
        """Set bounce to False if not specified."""
        if self.bounce is None:
            self.bounce = False
        return self.bounce

    def _prepare_reverse(self) -> bool:
        """Set reverse to False if not specified."""
        if self.reverse is None:
            self.reverse = False
        return self.reverse

    def _prepare_monochrome(self) -> bool:
        """Set monochrome to False if not specified."""
        if self.monochrome is None:
            self.monochrome = False
        return self.monochrome

    def _prepare_volume(self) -> float:
        """Set volume to 1.0 if not specified."""
        if self.volume is None:
            self.volume = 1.0
        return self.volume

    def _prepare_normalize(self) -> bool:
        """Set normalize to False if not specified."""
        if self.normalize is None:
            self.normalize = False
        return self.normalize


@dataclass
class PreparedFramesRequest(PreparedBaseRequest):
    """
    Prepares and stores 'frames' extraction request arguments.

    Fields
    -----
        `video` (Video):
            The object containing video metadata used to prepare the request.
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `image_format` (Optional[str]):
            Set the extracted image file format. Defaults to 'jpg' if not specified.
            See: `videoxt.constants.SUPPORTED_IMAGE_FORMATS`.
        `capture_rate` (Optional[int]):
            Capture every Nth video frame. Defaults to 1 if not specified, which
            extracts every frame within the extraction range.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the images. Defaults
            to the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the images by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the images by `n` degrees. Allowed values: 0, 90, 180 or 270.
            Defaults to 0 if not specified (no change).
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the images. Defaults to False if
            not specified.
        `images_expected` (int):
            The number of images expected to be written to disk. Not initialized.

    Public Methods
    -----
        - `prepare()` -> `PreparedFramesRequest`:
            Prepare fields and return a fully prepared frames request, required by a
            `FramesExtractor`.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared frames request.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    image_format: Optional[str] = None
    capture_rate: Optional[int] = None
    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    monochrome: Optional[bool] = None
    images_expected: int = field(init=False)

    def prepare(self) -> "PreparedFramesRequest":
        """
        Prepare request fields and return a `PreparedFramesRequest`, required by a
        `FramesExtractor`.
        """
        super().prepare()
        self._prepare_destpath()
        self._prepare_filename()
        self._prepare_image_format()
        self._prepare_capture_rate()
        self._prepare_resize()
        self._prepare_dimensions()
        self._prepare_rotate()
        self._prepare_monochrome()
        self._prepare_images_expected()
        self._is_prepared = True
        return self

    def _prepare_destpath(self) -> Path:
        """Set the directory the extracted images will be saved to."""
        self.destpath = P.prepare_destpath_frames(
            self.video.filepath, self.destdir, self.overwrite
        )
        return self.destpath

    def _prepare_filename(self) -> str:
        """Set filename to the video filename. Used for naming the extracted images."""
        if self.filename is None:
            self.filename = self.video.filepath.stem
        return self.filename

    def _prepare_image_format(self) -> str:
        """Set image format to 'jpg' if not specified."""
        if self.image_format is None:
            self.image_format = "jpg"
        return self.image_format

    def _prepare_capture_rate(self) -> int:
        """Set capture rate to 1 if not specified."""
        if self.capture_rate is None:
            self.capture_rate = 1
        return self.capture_rate

    def _prepare_resize(self) -> float:
        """Set resize to 1.0 if not specified."""
        if self.resize is None:
            self.resize = 1.0
        return self.resize

    def _prepare_dimensions(self) -> tuple[int, int]:
        """Set dimensions to the video dimensions if not specified."""
        self.dimensions = P.prepare_dimensions(
            self.video.dimensions, self.dimensions, self.resize
        )
        return self.dimensions

    def _prepare_rotate(self) -> int:
        """Set rotate to 0 if not specified."""
        if self.rotate is None:
            self.rotate = 0
        return self.rotate

    def _prepare_monochrome(self) -> bool:
        """Set monochrome to False if not specified."""
        if self.monochrome is None:
            self.monochrome = False
        return self.monochrome

    def _prepare_images_expected(self) -> int:
        """Set the number of images (frames) expected to be written to disk."""
        self.images_expected = P.prepare_images_expected(
            self.extraction_range.get("start_frame", None),
            self.extraction_range.get("stop_frame", None),
            self.capture_rate,
        )

        return self.images_expected


@dataclass
class PreparedGifRequest(PreparedBaseRequest):
    """
    Prepares and stores 'gif' extraction request arguments.

    Fields
    -----
        `video` (Video):
            The object containing video metadata used to prepare the request.
        `start_time` (Optional[float | int | str]):
            Specify the extraction's start time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to 0 if not specified.
        `stop_time` (Optional[float | int | str]):
            Specify the extraction's stop time in seconds, or as a string in "HH:MM:SS"
            format. Defaults to the video duration if not specified.
        `destdir` (Optional[pathlib.Path]):
            Specify the directory you want to save output to. Defaults to the video's
            directory if not specified.
        `filename` (Optional[str]):
            Specify the name of the extracted file(s). Defaults to the video filename
            if not specified.
        `verbose` (Optional[bool]):
            If True, the prepared request and extraction results will be printed as JSON
            to console. Defaults to False if not specified.
        `overwrite` (Optional[bool]):
            If True, permits overwriting the destination path if the file or directory
            already exists. Defaults to False if not specified.
        `fps` (Optional[float]):
            Override the frames per second (fps) value obtained from `cv2` when reading
            the video. This value is used to set the start and stop frames for the
            extraction range. This option should be used only in rare cases where `cv2`
            fails to accurately read the fps. If not specified, it defaults to the fps
            of the video as read by `cv2`.
        `dimensions` (Optional[Tuple[int, int]]):
            Specify the dimensions (frame width, frame height) of the gif. Defaults to
            the video dimensions if not specified.
        `resize` (Optional[float]):
            Resize the dimensions of the gif by a factor of `n`. A value of 0.5
            will halve the dimensions. If you specify `dimensions`, `resize` will apply
            to the dimensions you specify. Defaults to 1.0 if not specified (no change).
        `rotate` (Optional[int]):
            Rotate the gif by `n` degrees. Allowed values: 0, 90, 180 or 270. Defaults
            to 0 if not specified (no change).
        `speed` (Optional[float]):
            Set the speed of the extracted gif. A value of 0.5 will halve the playback
            speed of the gif. Defaults to 1.0 if not specified (no change).
        `bounce` (Optional[bool]):
            If True, bounce the extracted gif bommerang-style. Defaults to False if
            not specified.
        `reverse` (Optional[bool]):
            If True, reverse the extracted gif. Defaults to False if not specified.
        `monochrome` (Optional[bool]):
            If True, apply a black-and-white filter to the gif. Defaults to False if
            not specified.

    Public Methods
    -----
        - `prepare()` -> `PreparedGifRequest`:
            Prepare fields and return a fully prepared gif request, required by a
            `GifExtractor`.
        - `json()` -> `str`:
            Return a JSON string representation of the prepared gif request.
        - `verbose_print()` -> `None`:
            Print the JSON string to console with a title, without private keys.

    Properties:
    -----
        - `is_prepared` -> `bool`: If the request has been prepared.
        - `is_verbose` -> `bool | None`: Return the state of the verbose field.
    """

    dimensions: Optional[tuple[int, int]] = None
    resize: Optional[float] = None
    rotate: Optional[int] = None
    speed: Optional[float] = None
    bounce: Optional[bool] = None
    reverse: Optional[bool] = None
    monochrome: Optional[bool] = None

    def prepare(self) -> "PreparedGifRequest":
        """
        Prepare request fields and return a `PreparedGifRequest`, required by a
        `GifExtractor`.
        """
        super().prepare()
        self._prepare_destpath()
        self._prepare_resize()
        self._prepare_dimensions()
        self._prepare_rotate()
        self._prepare_speed()
        self._prepare_bounce()
        self._prepare_monochrome()
        self._prepare_reverse()
        self._is_prepared = True
        return self

    def _prepare_destpath(self) -> Path:
        """Set the destination path the extracted clip will be saved to."""
        self.destpath = P.prepare_destpath(
            self.video.filepath,
            self.filename,
            self.destdir,
            ".gif",
            self.overwrite,
        )

        return self.destpath

    def _prepare_resize(self) -> float:
        """Set resize to 1.0 if not specified."""
        if self.resize is None:
            self.resize = 1.0
        return self.resize

    def _prepare_dimensions(self) -> tuple[int, int]:
        """Set dimensions to the video dimensions if not specified."""
        self.dimensions = P.prepare_dimensions(
            self.video.dimensions, self.dimensions, self.resize
        )
        return self.dimensions

    def _prepare_rotate(self) -> int:
        """Set rotate to 0 if not specified."""
        if self.rotate is None:
            self.rotate = 0
        return self.rotate

    def _prepare_speed(self) -> float:
        """Set speed to 1.0 if not specified."""
        if self.speed is None:
            self.speed = 1.0
        return self.speed

    def _prepare_bounce(self) -> bool:
        """Set bounce to False if not specified."""
        if self.bounce is None:
            self.bounce = False
        return self.bounce

    def _prepare_monochrome(self) -> bool:
        """Set monochrome to False if not specified."""
        if self.monochrome is None:
            self.monochrome = False
        return self.monochrome

    def _prepare_reverse(self) -> bool:
        """Set reverse to False if not specified."""
        if self.reverse is None:
            self.reverse = False
        return self.reverse



import pytest

from videoxt.exceptions import (
    AudioWriteError,
    BuildImagePathError,
    ClipWriteError,
    ClosedVideoCaptureError,
    FrameReadError,
    FrameWriteError,
    GifWriteError,
    InvalidExtractionMethod,
    NoAudioError,
    PreparationError,
    ValidationError,
    VideoCaptureSetError,
    VideoValidationError,
    VideoXTError,
)


@pytest.mark.parametrize(
    "exception",
    [
        AudioWriteError,
        BuildImagePathError,
        ClipWriteError,
        ClosedVideoCaptureError,
        FrameReadError,
        FrameWriteError,
        GifWriteError,
        InvalidExtractionMethod,
        NoAudioError,
        PreparationError,
        ValidationError,
        VideoCaptureSetError,
        VideoValidationError,
        VideoXTError,
    ],
)
def test_errors(exception):
    with pytest.raises(exception) as exc_info:
        raise exception("Custom error message")

    assert str(exc_info.value) == "Custom error message"



"""Contains functions, objects that prepare shared request values for extraction."""
import math
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

import videoxt.utils as U
import videoxt.validators as V
from videoxt.exceptions import PreparationError


@dataclass
class ExtractionRange:
    """
    Container for variations of the user's requested extraction range.

    The range is represented in seconds, as timestamps, and frame numbers. Validations
    and preparations are run on init.

    Fields:
    -----
        `duration_seconds` (float):
            Duration of the video in seconds.
        `frame_count` (int):
            Number of frames in the video.
        `start_time` (float | int | str):
            Start time of the extraction in seconds or as a timestamp (`HH:MM:SS`).
        `stop_time` (float | int | str):
            Stop time of the extraction in seconds or as a timestamp (`HH:MM:SS`).
        `fps` (float):
            Frames per second to use for extraction.

    Usage:
    -----
    >>> from videoxt.preppers import ExtractionRange
    >>> extraction_range = ExtractionRange(
    ...     duration_seconds=20,
    ...     frame_count=600,
    ...     start_time=10,
    ...     stop_time=0,
    ...     fps=30,
    ... )
    >>> extraction_range.start_second
    0
    >>> extraction_range.stop_second
    10
    >>> extraction_range.start_timestamp
    '0:00:00'
    >>> extraction_range.stop_timestamp
    '0:00:10'
    >>> extraction_range.start_frame
    0
    >>> extraction_range.stop_frame
    300
    """

    duration_seconds: float = field(repr=False)
    frame_count: int = field(repr=False)
    start_time: float | int | str
    stop_time: float | int | str
    fps: float = field(repr=False)
    start_second: float = field(init=False)
    stop_second: float = field(init=False)
    start_timestamp: str = field(init=False)
    stop_timestamp: str = field(init=False)
    start_frame: int = field(init=False)
    stop_frame: int = field(init=False)

    def __post_init__(self):
        """
        Start procedure that prepares and validates the extraction range.

        - Convert the requested start and stop times to seconds (float)
        - Validate the range against the video's duration
        - Convert each request second to timestamps (str) and frame numbers (int).
        """
        self._prepare_start_second()
        self._prepare_stop_second()
        self._validate_range()
        self._prepare_start_timestamp()
        self._prepare_stop_timestamp()
        self._prepare_start_frame()
        self._prepare_stop_frame()

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary of the validated extraction start and stop points."""
        return {
            "start_second": self.start_second,
            "stop_second": self.stop_second,
            "start_timestamp": self.start_timestamp,
            "stop_timestamp": self.stop_timestamp,
            "start_frame": self.start_frame,
            "stop_frame": self.stop_frame,
        }

    def _prepare_start_second(self) -> float:
        """Convert the start time requested to seconds (float)."""
        self.start_second = (
            float(self.start_time)
            if isinstance(self.start_time, (float, int))
            else U.timestamp_to_seconds(self.start_time)
        )
        return self.start_second

    def _prepare_stop_second(self) -> float:
        """Convert the stop time requested to seconds (float)."""
        self.stop_second = (
            float(self.stop_time)
            if isinstance(self.stop_time, (float, int))
            else U.timestamp_to_seconds(self.stop_time)
        )

        if self.stop_second > self.duration_seconds:
            self.stop_second = self.duration_seconds

        return self.stop_second

    def _validate_range(self) -> tuple[float, float, float]:
        """
        Validate start and stop seconds against the video's duration and each other.

        Returns:
        -----
            `tuple[float, float, float]`: Validated (start, stop, duration) in seconds.
        """
        start, stop, duration = V.valid_extraction_range(
            self.start_second, self.stop_second, self.duration_seconds
        )
        return start, stop, duration

    def _prepare_start_timestamp(self) -> str:
        """Convert the start time requested to a timestamp (str) `HH:MM:SS`."""
        self.start_timestamp = U.seconds_to_timestamp(self.start_second)
        return self.start_timestamp

    def _prepare_stop_timestamp(self) -> str:
        """Convert the stop time requested to a timestamp (str) `HH:MM:SS`."""
        self.stop_timestamp = U.seconds_to_timestamp(self.stop_second)
        return self.stop_timestamp

    def _prepare_start_frame(self) -> int:
        """Convert the start time requested to a frame number (int)."""
        self.start_frame = math.floor(self.start_second * self.fps)
        return self.start_frame

    def _prepare_stop_frame(self) -> int:
        """Convert the stop time requested to a frame number (int)."""
        self.stop_frame = (
            self.frame_count
            if self.stop_second is None
            else math.floor(self.stop_second * self.fps)
        )
        return self.stop_frame


def prepare_extraction_range(
    duration_seconds: float,
    frame_count: int,
    prepared_start_time: Optional[float | int | str] = None,
    prepared_stop_time: Optional[float | int | str] = None,
    prepared_fps: Optional[float] = None,
) -> dict[str, Any]:
    """
    Return a dictionary representing a validated extraction start and stop points.

    The dictionary contains variations of the requested start and stop times as seconds
    (float), timestamps (str), and frame numbers (int); including the requested values.

    Args:
    -----
        `duration_seconds` (float):
            Duration of the video in seconds.
        `frame_count` (int):
            Number of frames in the video.
        `prepared_start_time` (Optional[float | int | str]):
            Prepared start time of the extraction in seconds or as a timestamp
            (`HH:MM:SS`).
        `request_stop_time` (Optional[float | int | str]):
            Prepared stop time of the extraction in seconds or as a timestamp
            (`HH:MM:SS`).
        `prepared_fps` (Optional[float]):
            The prepared frames per second used to calculate the frame numbers.

    Returns:
    -----
        `dict[str, Any]`: Dictionary of validated extraction start and stop points.

    Raises:
    -----
        `PreparationError`: If the prepared start and stop times or fps are None.
    """
    if (
        prepared_start_time is None
        or prepared_stop_time is None
        or prepared_fps is None
    ):
        raise PreparationError("Start time, stop time or fps are None.")

    extraction_range = ExtractionRange(
        duration_seconds=duration_seconds,
        frame_count=frame_count,
        start_time=prepared_start_time,
        stop_time=prepared_stop_time,
        fps=prepared_fps,
    )

    return extraction_range.to_dict()


def prepare_dimensions(
    video_dimensions: tuple[int, int],
    request_dimensions: Optional[tuple[int, int]] = None,
    request_resize: Optional[float] = None,
) -> tuple[int, int]:
    """
    Return dimensions to use for extraction.

    Args:
    -----
        `video_dimensions` (tuple[int, int]):
            Dimensions of the video.
        `request_dimensions` (Optional[tuple[int, int]]):
            Optional dimensions that override the video's dimensions.
        `request_resize` (Optional[float]):
            Requested resize factor to apply to the video's dimensions.

    Returns:
    -----
        `tuple[int, int]`: Dimensions to use for extraction.
    """
    dims = request_dimensions or video_dimensions

    if request_resize != 1.0 and request_resize is not None:
        dims = (int(dims[0] * request_resize), int(dims[1] * request_resize))

    return dims


def prepare_destpath(
    video_filepath: Path,
    request_filename: Optional[str] = None,
    request_destdir: Optional[Path] = None,
    prepared_suffix: Optional[str] = None,
    prepared_overwrite: Optional[bool] = None,
) -> Path:
    """
    Construct and return the destination path for the extracted file.

    Args:
    -----
        `video_filepath` (pathlib.Path):
            Path to the video file.
        `request_filename` (Optional[str]):
            Optional filename to use. Defaults to None.
        `request_destdir` (Optional[pathlib.Path]):
            Optional directory to save the processed file. Defaults to None.
        `prepared_suffix` (str):
            Required file suffix to use (with or without the leading '.').
        `prepared_overwrite` (bool):
            If True, allows overwriting an existing file with the same name.
            Defaults to False.

    Returns:
    -----
        `pathlib.Path`: Path to the destination file.

    Raises:
    -----
        `PreparationError`: If the prepared suffix is None.

    Notes:
    -----
        - Uses `request_filename` with the specified `suffix` if provided; otherwise,
            uses video file's name.
        - Saves the file in `request_destdir` if provided; otherwise, saves in the
            video's directory.
        - Handles overwriting: if `overwrite` is True, may overwrite existing file with
            the same name; if False, ensures a unique filename before saving.
    """
    if prepared_suffix is None:
        raise PreparationError("Suffix is None.")

    base_dir = request_destdir or video_filepath.parent
    suffix = (
        prepared_suffix if prepared_suffix.startswith(".") else f".{prepared_suffix}"
    )
    filename = f"{request_filename or video_filepath.stem}{suffix}"
    dest_path = base_dir / filename

    if prepared_overwrite is True and dest_path != video_filepath:
        return dest_path

    return U.enumerate_filepath(dest_path, tag="_vxt")


def prepare_start_time(
    request_start_time: Optional[float | int | str] = None,
) -> float | int | str:
    """
    Return the start time requested or 0 if not provided.

    Args:
    -----
        `request_start_time` (Optional[float | int | str]):
            Optional start time requested. Defaults to None.

    Returns:
    -----
        `float | int | str`: The start time requested or 0 if not provided.
    """
    return request_start_time if request_start_time is not None else 0


def prepare_stop_time(
    video_duration_seconds: float, request_stop_time: Optional[float | int | str] = None
) -> float | int | str:
    """
    Return the stop time requested or the video's duration if not provided.

    Args:
    -----
        `video_duration_seconds` (float):
            Duration of the video in seconds.
        `request_stop_time` (Optional[float | int | str]):
            Optional stop time requested. Defaults to None.

    Returns:
    -----
        `float | int | str`: The stop time requested or the video's duration if not
            provided.
    """
    return (
        request_stop_time if request_stop_time is not None else video_duration_seconds
    )


def prepare_fps(video_fps: float, request_fps: Optional[float] = None) -> float:
    """
    Return the frames per second requested or the video's fps if not provided.

    Args:
    -----
        `video_fps` (float):
            Frames per second of the video.
        `request_fps` (Optional[float]):
            Optional frames per second requested. Defaults to None.

    Returns:
    -----
        `float`: The frames per second requested or the video's fps if not provided.
    """
    return request_fps if request_fps is not None else video_fps


def prepare_destpath_frames(
    video_filepath: Path,
    request_destdir: Optional[Path] = None,
    prepared_overwrite: Optional[bool] = None,
) -> Path:
    """
    Return the directory extracted images will be saved in.

    If no specific `destdir` is requested, the images will be saved in a directory
    created in the same location as the video file. The directory will share the
    video file's name and have a '_frames' suffix.

    Returns:
    -----
        `pathlib.Path`: The directory for saving images extracted from the video.
    """
    if request_destdir is not None and request_destdir != video_filepath:
        return request_destdir

    default_destdir = video_filepath.parent / video_filepath.name

    if prepared_overwrite is True:
        return default_destdir.with_name(f"{video_filepath.name}_frames")

    return U.enumerate_dir(default_destdir, tag="_frames")


def prepare_images_expected(
    start_frame: Optional[int] = None,
    stop_frame: Optional[int] = None,
    capture_rate: Optional[int] = None,
) -> int:
    """
    Return the number of images (frames) expected to be written to disk.

    Args:
    -----
        `start_frame` (Optional[int]):
            The frame number to start extraction from.
        `stop_frame` (Optional[int]):
            The frame number to stop extraction at.
        `capture_rate` (Optional[int]):
            The number of frames to increment by.

    Returns:
    -----
        `int`: The number of images expected to be extracted.

    Raises:
    -----
        `PreparationError`:
        - If the start frame, stop frame or capture rate is None
        - If the capture rate is zero.
    """
    if start_frame is None or stop_frame is None or capture_rate is None:
        raise PreparationError("Start frame, stop frame or capture rate is None.")

    try:
        return math.ceil((stop_frame - start_frame) / capture_rate)
    except ZeroDivisionError as e:
        raise PreparationError("Capture rate is zero.") from e



"""Contains Video class and functions for validating and retrieving video properties."""
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path
from typing import Any

import cv2  # type: ignore
from moviepy.editor import VideoFileClip  # type: ignore

import videoxt.utils as U
import videoxt.validators as V
from videoxt.exceptions import (
    ClosedVideoCaptureError,
    ValidationError,
    VideoValidationError,
)


@dataclass
class Video:
    """
    Validate, set and store video properties such as the file path, dimensions, fps,
    frame count, duration, filesize, and whether the video has audio. A validated
    `Video` object is required to validate and calculate attributes in primary objects
    and functions throughout.

    Args:
    -----
        `filepath` (pathlib.Path): Path to the video file.

    Attributes:
    -----
        `dimensions` (tuple[int, int]):
            Dimensions of the video as (width, height).
        `fps` (float):
            Frame rate of the video.
        `frame_count` (int):
            Number of frames in the video.
        `duration` (datetime.timedelta):
            Duration of the video as a timedelta object.
        `duration_seconds` (float):
            Duration of the video in seconds.
        `duration_timestamp` (str):
            Duration of the video in `HH:MM:SS` format.
        `has_audio` (bool):
            True if the video has audio, False otherwise.
        `filesize_bytes` (int):
            Size of the video file in bytes.
        `filesize` (str):
            Size of the video file in a human-readable format.
    """

    filepath: Path
    dimensions: tuple[int, int] = field(init=False)
    fps: float = field(init=False)
    frame_count: int = field(init=False)
    duration: timedelta = field(init=False)
    duration_seconds: float = field(init=False)
    duration_timestamp: str = field(init=False)
    has_audio: bool = field(init=False)
    filesize_bytes: int = field(init=False)
    filesize: str = field(init=False)

    def __post_init__(self) -> None:
        """Run video file validations and set attributes read from a validated video."""
        self.validate_filepath()
        self.validate_filesize_bytes()
        self.setattrs_from_opencap()
        self.validate_dimensions()
        self.validate_fps()
        self.validate_frame_count()
        self.setattrs_duration()
        self.setattr_filesize()
        self.setattr_has_audio()

    def validate_filepath(self) -> Path:
        """Validate the file is a video file and exists."""
        self.filepath = V.valid_filepath(self.filepath, is_video=True)
        return self.filepath

    def setattr_filesize(self) -> str:
        """Set the filesize attribute to a human-readable format."""
        self.filesize = U.convert_bytes(self.filesize_bytes)
        return self.filesize

    def setattrs_from_opencap(self) -> tuple[tuple[int, int], float, int]:
        """Set the dimensions, fps and frame count read from an opened video capture."""
        properties = fetch_video_properties(self.filepath)
        self.dimensions = properties.get("dimensions", None)
        self.fps = properties.get("fps", None)
        self.frame_count = properties.get("frame_count", None)

        return self.dimensions, self.fps, self.frame_count

    def validate_dimensions(self) -> tuple[int, int]:
        """Validate the video frame width and height are positive integers."""
        if self.dimensions is None:
            raise VideoValidationError(
                "The video dimensions could not be read from the video file. Please "
                "check the file."
            )

        try:
            self.dimensions = V.valid_dimensions(self.dimensions)
        except ValidationError as err:
            raise VideoValidationError(
                "The video dimensions are invalid. Please check the file."
            ) from err

        return self.dimensions

    def validate_fps(self) -> float:
        """Validate the video fps is a positive float."""
        if self.fps is None:
            raise VideoValidationError(
                "The video fps was read as None from the video file. Please check the "
                "file."
            )

        try:
            self.fps = V.positive_float(self.fps)
        except ValidationError as err:
            raise VideoValidationError(
                "The video fps was read as a non-positive float from the video file. "
                "Please check the file."
            ) from err

        return self.fps

    def validate_frame_count(self) -> int:
        """Validate the video frame count is a positive integer."""
        if self.frame_count is None:
            raise VideoValidationError(
                "The video frame count could not be read from the video file. Please "
                "check the file."
            )

        try:
            self.frame_count = V.positive_int(self.frame_count)
        except ValidationError as err:
            raise VideoValidationError(
                "The video frame count was read as a non-positive integer from the "
                "video file. Please check the file."
            ) from err

        return self.frame_count

    def setattrs_duration(self) -> tuple[timedelta, float, str]:
        """Set the duration attributes."""
        self.duration = U.calculate_duration(self.frame_count, self.fps)
        self.duration_seconds = self.duration.total_seconds()
        self.duration_timestamp = U.seconds_to_timestamp(self.duration_seconds)
        return self.duration, self.duration_seconds, self.duration_timestamp

    def validate_filesize_bytes(self) -> int:
        """Validate the video file size in bytes is a positive integer."""
        try:
            self.filesize_bytes = V.positive_int(self.filepath.stat().st_size)
        except ValidationError as err:
            raise VideoValidationError(
                "The video file size could not be read. Please check the file."
            ) from err

        return self.filesize_bytes

    def setattr_has_audio(self) -> bool:
        """Set the has_audio attribute with moviepy.editor import VideoFileClip.audio"""
        with VideoFileClip(str(self.filepath)) as clip:
            self.has_audio = bool(clip.audio)
        return self.has_audio


def fetch_video_properties(filepath: Path) -> dict[str, Any]:
    """
    Open the video file to retrieve and return the video's dimensions, fps, and frame
    count as a dictionary.

    Args:
    -----
        `filepath` (pathlib.Path): Path to the video file.

    Returns:
    -----
        `dict[str, Any]`:
            Dictionary containing unvalidated video properties:

            - "dimensions" (tuple): Dimensions of the video as (width, height).
            - "fps" (float): Frame rate of the video.
            - "frame_count" (int): Number of frames in the video.
    """
    with open_video_capture(filepath) as opencap:
        frame_height: int = opencap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        frame_width: int = opencap.get(cv2.CAP_PROP_FRAME_WIDTH)
        fps: float = opencap.get(cv2.CAP_PROP_FPS)
        frame_count: int = opencap.get(cv2.CAP_PROP_FRAME_COUNT)

    return {
        "dimensions": (frame_width, frame_height),
        "fps": fps,
        "frame_count": frame_count,
    }


@contextmanager
def open_video_capture(filepath: Path) -> Iterator[cv2.VideoCapture]:
    """
    Context manager for opening a video file with `cv2.VideoCapture`.

    Usage:
    -----
        >>> from videoxt.video import open_video_capture
        >>> with open_video_capture('path/to/video.mp4') as opencap:
        ...     # do something with 'opencap'
        ...     type(opencap)
        <class 'cv2.VideoCapture'>

    Args:
    -----
        `filepath` (pathlib.Path): Path to the video file.

    Yields:
    -----
        `cv2.VideoCapture`: An opened cv2 video capture.

    Raises:
    -----
        - `cv2.error`:
            - If `cv2.VideoCapture().isOpened()` returns False.
            - If the file is not supported by OpenCV.
            - For more details see:
            https://docs.opencv.org/4.8.0/d8/dfe/classcv_1_1VideoCapture.html
        - `UnboundLocalError`: If the user cancels the operation. This is raised by
            `cv2.VideoCapture.release()` when the video capture is not opened.
    """
    try:
        video_capture = cv2.VideoCapture(str(filepath))
        if not video_capture.isOpened():
            raise ClosedVideoCaptureError(
                "Unable to open the video file or maintain it in an open state. Please "
                "check the file.\ncv2.VideoCapture().isOpened() returned False."
            )

    except cv2.error as err:
        raise ClosedVideoCaptureError(
            "A cv2.error occurred opening the video file with cv2.VideoCapture. "
            f"Please check the file, got {filepath!r}"
        ) from err

    else:
        yield video_capture

    finally:
        try:
            video_capture.release()
        except UnboundLocalError:
            pass  # Ctrl+C keyboard interrupt
